import numpy as np


def clip_boxes(boxes, reference_box, copy=True):
    """Clip boxes to reference box.
    
    References:
        `clip_to_window` in TensorFlow object detection API.
    """
    if copy:
        boxes = boxes.copy()
    ref_x_min, ref_y_min, ref_x_max, ref_y_max = reference_box[:4]
    lower = np.array([ref_x_min, ref_y_min, ref_x_min, ref_y_min])
    upper = np.array([ref_x_max, ref_y_max, ref_x_max, ref_y_max])
    np.clip(boxes[..., :4], lower, upper, boxes[..., :4])
    return boxes
    
    
def clip_boxes_to_image(boxes, image_width, image_height, subpixel=True, copy=True):
    """Clip boxes to image boundaries.
    
    References:
        `clip_boxes` in py-faster-rcnn
        `core.boxes_op_list.clip_to_window` in TensorFlow object detection API.
        `structures.Boxes.clip` in detectron2
        
    Notes:
        Equivalent to `clip_boxes(boxes, [0,0,image_width-1,image_height-1], copy)`
    """
    if not subpixel:
        image_width -= 1
        image_height -= 1
    reference_box = [0, 0, image_width, image_height]
    return clip_boxes(boxes, reference_box, copy)



import copy
import dataclasses
import json
import numbers
import os
import warnings
import xml.etree.ElementTree as ET
from collections import OrderedDict
from dataclasses import dataclass, field
from typing import Callable, Dict, List, Optional, Union

import lxml
import lxml.builder
import numpy as np

import khandy

__all__ = ['DetectIrObject', 'DetectIrRecord', 'load_detect',
           'save_detect', 'convert_detect', 'replace_detect_label',
           'load_coco_class_names', 'crop_detect', 'fliter_detect', 
           'fliter_detect_by_label']


@dataclass
class DetectIrObject:
    """Intermediate Representation Format of Object
    """
    label: str
    x_min: float
    y_min: float
    x_max: float
    y_max: float

    def __post_init__(self):
        if self.x_min > self.x_max:
            warnings.warn("Warning: x_min is greater than x_max, swapping values.")
            self.x_min, self.x_max = self.x_max, self.x_min
        if self.y_min > self.y_max:
            warnings.warn("Warning: y_min is greater than y_max, swapping values.")
            self.y_min, self.y_max = self.y_max, self.y_min
    
    @property
    def area(self) -> float:
        """Calculate the area of the bounding box."""
        return (self.x_max - self.x_min) * (self.y_max - self.y_min)


@dataclass
class DetectIrRecord:
    """Intermediate Representation Format of Record
    """
    filename: str
    width: int
    height: int
    objects: List[DetectIrObject] = field(default_factory=list)


@dataclass
class PascalVocSource:
    database: str = ''
    annotation: str = ''
    image: str = ''


@dataclass
class PascalVocSize:
    height: int
    width: int
    depth: int


@dataclass
class PascalVocBndbox:
    xmin: float
    ymin: float
    xmax: float
    ymax: float


@dataclass
class PascalVocObject:
    name: str
    pose: str = 'Unspecified'
    truncated: int = 0
    difficult: int = 0
    bndbox: Optional[PascalVocBndbox] = None


@dataclass
class PascalVocRecord:
    folder: str = ''
    filename: str = ''
    path: str = ''
    source: PascalVocSource = field(default_factory=PascalVocSource)
    size: Optional[PascalVocSize] = None
    segmented: int = 0
    objects: List[PascalVocObject] = field(default_factory=list)


class PascalVocHandler:
    @staticmethod
    def load(filename, **kwargs) -> PascalVocRecord:
        pascal_voc_record = PascalVocRecord()

        xml_tree = ET.parse(filename)
        pascal_voc_record.folder = xml_tree.find('folder').text
        pascal_voc_record.filename = xml_tree.find('filename').text
        pascal_voc_record.path = xml_tree.find('path').text
        pascal_voc_record.segmented = xml_tree.find('segmented').text

        source_tag = xml_tree.find('source')
        pascal_voc_record.source = PascalVocSource(
            database=source_tag.find('database').text,
            # annotation=source_tag.find('annotation').text,
            # image=source_tag.find('image').text
        )

        size_tag = xml_tree.find('size')
        pascal_voc_record.size = PascalVocSize(
            width=int(size_tag.find('width').text),
            height=int(size_tag.find('height').text),
            depth=int(size_tag.find('depth').text)
        )

        object_tags = xml_tree.findall('object')
        for index, object_tag in enumerate(object_tags):
            bndbox_tag = object_tag.find('bndbox')
            bndbox = PascalVocBndbox(
                xmin=float(bndbox_tag.find('xmin').text),
                ymin=float(bndbox_tag.find('ymin').text),
                xmax=float(bndbox_tag.find('xmax').text),
                ymax=float(bndbox_tag.find('ymax').text)
            )
            pascal_voc_object = PascalVocObject(
                name=object_tag.find('name').text,
                pose=object_tag.find('pose').text,
                truncated=object_tag.find('truncated').text,
                difficult=object_tag.find('difficult').text,
                bndbox=bndbox
            )
            pascal_voc_record.objects.append(pascal_voc_object)
        return pascal_voc_record

    @staticmethod
    def save(filename, pascal_voc_record: PascalVocRecord):
        maker = lxml.builder.ElementMaker()
        xml = maker.annotation(
            maker.folder(pascal_voc_record.folder),
            maker.filename(pascal_voc_record.filename),
            maker.path(pascal_voc_record.path),
            maker.source(
                maker.database(pascal_voc_record.source.database),
            ),
            maker.size(
                maker.width(str(pascal_voc_record.size.width)),
                maker.height(str(pascal_voc_record.size.height)),
                maker.depth(str(pascal_voc_record.size.depth)),
            ),
            maker.segmented(str(pascal_voc_record.segmented)),
        )

        for pascal_voc_object in pascal_voc_record.objects:
            object_tag = maker.object(
                maker.name(pascal_voc_object.name),
                maker.pose(pascal_voc_object.pose),
                maker.truncated(str(pascal_voc_object.truncated)),
                maker.difficult(str(pascal_voc_object.difficult)),
                maker.bndbox(
                    maker.xmin(str(float(pascal_voc_object.bndbox.xmin))),
                    maker.ymin(str(float(pascal_voc_object.bndbox.ymin))),
                    maker.xmax(str(float(pascal_voc_object.bndbox.xmax))),
                    maker.ymax(str(float(pascal_voc_object.bndbox.ymax))),
                ),
            )
            xml.append(object_tag)

        if not filename.endswith('.xml'):
            filename = filename + '.xml'
        with open(filename, 'wb') as f:
            f.write(lxml.etree.tostring(xml, pretty_print=True, encoding='utf-8'))

    @staticmethod
    def to_ir(pascal_voc_record: PascalVocRecord) -> DetectIrRecord:
        ir_record = DetectIrRecord(
            filename=pascal_voc_record.filename,
            width=pascal_voc_record.size.width,
            height=pascal_voc_record.size.height
        )
        for pascal_voc_object in pascal_voc_record.objects:
            ir_object = DetectIrObject(
                label=pascal_voc_object.name,
                x_min=pascal_voc_object.bndbox.xmin,
                y_min=pascal_voc_object.bndbox.ymin,
                x_max=pascal_voc_object.bndbox.xmax,
                y_max=pascal_voc_object.bndbox.ymax
            )
            ir_record.objects.append(ir_object)
        return ir_record

    @staticmethod
    def from_ir(ir_record: DetectIrRecord) -> PascalVocRecord:
        pascal_voc_record = PascalVocRecord(
            filename=ir_record.filename,
            size=PascalVocSize(
                width=ir_record.width,
                height=ir_record.height,
                depth=3
            )
        )
        for ir_object in ir_record.objects:
            pascal_voc_object = PascalVocObject(
                name=ir_object.label,
                bndbox=PascalVocBndbox(
                    xmin=ir_object.x_min,
                    ymin=ir_object.y_min,
                    xmax=ir_object.x_max,
                    ymax=ir_object.y_max,
                )
            )
            pascal_voc_record.objects.append(pascal_voc_object)
        return pascal_voc_record


class _NumpyEncoder(json.JSONEncoder):
    """ Special json encoder for numpy types """
    def default(self, obj):
        if isinstance(obj, (np.bool_,)):
            return bool(obj)
        elif isinstance(obj, (np.int_, np.intc, np.intp, np.int8,
                              np.int16, np.int32, np.int64, np.uint8,
                              np.uint16, np.uint32, np.uint64)):
            return int(obj)
        elif isinstance(obj, (np.float_, np.float16, np.float32,
                              np.float64)):
            return float(obj)
        elif isinstance(obj, (np.ndarray,)):
            return obj.tolist()
        return json.JSONEncoder.default(self, obj)


@dataclass
class LabelmeShape:
    label: str
    points: np.ndarray
    shape_type: str
    flags: Dict[str, bool] = field(default_factory=dict)
    group_id: Optional[int] = None
    description: str = ''
    mask: Optional[np.ndarray] = None

    def __post_init__(self):
        self.points = np.asarray(self.points)


@dataclass
class LabelmeRecord:
    version: str = '4.5.6'
    flags: Dict[str, bool] = field(default_factory=dict)
    shapes: List[LabelmeShape] = field(default_factory=list)
    imagePath: Optional[str] = None
    imageData: Optional[str] = None
    imageHeight: Optional[int] = None
    imageWidth: Optional[int] = None

    def __post_init__(self):
        for k, shape in enumerate(self.shapes):
            self.shapes[k] = LabelmeShape(**shape)


class LabelmeHandler:
    @staticmethod
    def load(filename, **kwargs) -> LabelmeRecord:
        json_content = khandy.load_json(filename)
        return LabelmeRecord(**json_content)

    @staticmethod
    def save(filename, labelme_record: LabelmeRecord):
        json_content = dataclasses.asdict(labelme_record)
        khandy.save_json(filename, json_content, cls=_NumpyEncoder)

    @staticmethod
    def to_ir(labelme_record: LabelmeRecord) -> DetectIrRecord:
        ir_record = DetectIrRecord(
            filename=labelme_record.imagePath,
            width=labelme_record.imageWidth,
            height=labelme_record.imageHeight
        )
        for labelme_shape in labelme_record.shapes:
            if labelme_shape.shape_type == 'polygon':
                ir_object = DetectIrObject(
                    label=labelme_shape.label,
                    x_min=np.min(labelme_shape.points[:, 0]),
                    y_min=np.min(labelme_shape.points[:, 1]),
                    x_max=np.max(labelme_shape.points[:, 0]),
                    y_max=np.max(labelme_shape.points[:, 1]),
                )
            elif labelme_shape.shape_type == 'rectangle':
                ir_object = DetectIrObject(
                    label=labelme_shape.label,
                    x_min=labelme_shape.points[0][0],
                    y_min=labelme_shape.points[0][1],
                    x_max=labelme_shape.points[1][0],
                    y_max=labelme_shape.points[1][1],
                )
            else:
                continue
            ir_record.objects.append(ir_object)
        return ir_record

    @staticmethod
    def from_ir(ir_record: DetectIrRecord) -> LabelmeRecord:
        labelme_record = LabelmeRecord(
            imagePath=ir_record.filename,
            imageWidth=ir_record.width,
            imageHeight=ir_record.height
        )
        for ir_object in ir_record.objects:
            labelme_shape = LabelmeShape(
                label=ir_object.label,
                shape_type='rectangle',
                points=[[ir_object.x_min, ir_object.y_min],
                        [ir_object.x_max, ir_object.y_max]]
            )
            labelme_record.shapes.append(labelme_shape)
        return labelme_record


@dataclass
class YoloObject:
    label: str
    x_center: float
    y_center: float
    width: float
    height: float


@dataclass
class YoloRecord:
    filename: Optional[str] = None
    width: Optional[int] = None
    height: Optional[int] = None
    objects: List[YoloObject] = field(default_factory=list)


class YoloHandler:
    @staticmethod
    def load(filename, **kwargs) -> YoloRecord:
        assert 'image_filename' in kwargs
        assert 'width' in kwargs and 'height' in kwargs

        records = khandy.load_list(filename)
        yolo_record = YoloRecord(
            filename=kwargs.get('image_filename'),
            width=kwargs.get('width'),
            height=kwargs.get('height'))
        for record in records:
            record_parts = record.split()
            yolo_record.objects.append(YoloObject(
                label=record_parts[0],
                x_center=float(record_parts[1]),
                y_center=float(record_parts[2]),
                width=float(record_parts[3]),
                height=float(record_parts[4]),
            ))
        return yolo_record

    @staticmethod
    def save(filename, yolo_record: YoloRecord):
        records = []
        for object in yolo_record.objects:
            records.append(f'{object.label} {object.x_center} {object.y_center} {object.width} {object.height}')
        if not filename.endswith('.txt'):
            filename = filename + '.txt'
        khandy.save_list(filename, records)

    @staticmethod
    def to_ir(yolo_record: YoloRecord) -> DetectIrRecord:
        ir_record = DetectIrRecord(
            filename=yolo_record.filename,
            width=yolo_record.width,
            height=yolo_record.height
        )
        for yolo_object in yolo_record.objects:
            x_min = (yolo_object.x_center - 0.5 * yolo_object.width) * yolo_record.width
            y_min = (yolo_object.y_center - 0.5 * yolo_object.height) * yolo_record.height
            x_max = (yolo_object.x_center + 0.5 * yolo_object.width) * yolo_record.width
            y_max = (yolo_object.y_center + 0.5 * yolo_object.height) * yolo_record.height
            ir_object = DetectIrObject(
                label=yolo_object.label,
                x_min=x_min,
                y_min=y_min,
                x_max=x_max,
                y_max=y_max
            )
            ir_record.objects.append(ir_object)
        return ir_record

    @staticmethod
    def from_ir(ir_record: DetectIrRecord) -> YoloRecord:
        yolo_record = YoloRecord(
            filename=ir_record.filename,
            width=ir_record.width,
            height=ir_record.height
        )
        for ir_object in ir_record.objects:
            x_center = (ir_object.x_max + ir_object.x_min) / (2 * ir_record.width)
            y_center = (ir_object.y_max + ir_object.y_min) / (2 * ir_record.height)
            width = abs(ir_object.x_max - ir_object.x_min) / ir_record.width
            height = abs(ir_object.y_max - ir_object.y_min) / ir_record.height
            yolo_object = YoloObject(
                label=ir_object.label,
                x_center=x_center,
                y_center=y_center,
                width=width,
                height=height,
            )
            yolo_record.objects.append(yolo_object)
        return yolo_record


@dataclass
class CocoObject:
    label: str
    x_min: float
    y_min: float
    width: float
    height: float


@dataclass
class CocoRecord:
    filename: str
    width: int
    height: int
    objects: List[CocoObject] = field(default_factory=list)


class CocoHandler:
    @staticmethod
    def load(filename, **kwargs) -> List[CocoRecord]:
        json_data = khandy.load_json(filename)

        images = json_data['images']
        annotations = json_data['annotations']
        categories = json_data['categories']

        label_map = {}
        for cat_item in categories:
            label_map[cat_item['id']] = cat_item['name']

        coco_records = OrderedDict()
        for image_item in images:
            coco_records[image_item['id']] = CocoRecord(
                filename=image_item['file_name'],
                width=image_item['width'],
                height=image_item['height'],
                objects=[])

        for annotation_item in annotations:
            coco_object = CocoObject(
                label=label_map[annotation_item['category_id']],
                x_min=annotation_item['bbox'][0],
                y_min=annotation_item['bbox'][1],
                width=annotation_item['bbox'][2],
                height=annotation_item['bbox'][3])
            coco_records[annotation_item['image_id']].objects.append(coco_object)
        return list(coco_records.values())

    @staticmethod
    def to_ir(coco_record: CocoRecord) -> DetectIrRecord:
        ir_record = DetectIrRecord(
            filename=coco_record.filename,
            width=coco_record.width,
            height=coco_record.height,
        )
        for coco_object in coco_record.objects:
            ir_object = DetectIrObject(
                label=coco_object.label,
                x_min=coco_object.x_min,
                y_min=coco_object.y_min,
                x_max=coco_object.x_min + coco_object.width,
                y_max=coco_object.y_min + coco_object.height
            )
            ir_record.objects.append(ir_object)
        return ir_record

    @staticmethod
    def from_ir(ir_record: DetectIrRecord) -> CocoRecord:
        coco_record = CocoRecord(
            filename=ir_record.filename,
            width=ir_record.width,
            height=ir_record.height
        )
        for ir_object in ir_record.objects:
            coco_object = CocoObject(
                label=ir_object.label,
                x_min=ir_object.x_min,
                y_min=ir_object.y_min,
                width=ir_object.x_max - ir_object.x_min,
                height=ir_object.y_max - ir_object.y_min
            )
            coco_record.objects.append(coco_object)
        return coco_record

    @staticmethod
    def load_class_names(filename):
        json_data = khandy.load_json(filename)
        categories = json_data['categories']
        return [cate_item['name'] for cate_item in categories]


@dataclass
class LabelBeeObject:
    id: str
    sourceID: str
    order: int
    textAttribute: str
    attribute: str
    x: float
    y: float
    width: float
    height: float
    valid: bool = True
    

@dataclass
class LabelBeeStep:
    toolName: str = 'rectTool'
    dataSourceStep: int = 0
    result: List[LabelBeeObject] = field(default_factory=list)
    
    def __post_init__(self):
        dst_result = []
        for item in self.result:
            new_item = {}
            for item_key in item:
                if item_key in LabelBeeObject.__annotations__:
                    new_item[item_key] = item[item_key]
            dst_result.append(LabelBeeObject(**new_item))
        self.result = dst_result
        # self.result = [LabelBeeObject(**item) for item in self.result]


@dataclass
class LabelBeeRecord:
    width: int
    height: int
    rotate: int = 0
    valid: bool = True
    step_1: LabelBeeStep = field(default_factory=LabelBeeStep)
    
    def __post_init__(self):
        if isinstance(self.step_1, dict):
            self.step_1 = LabelBeeStep(**self.step_1)


class LabelBeeHandler:
    """
    Notes: 
        Only support one step and its toolName must be `rectTool`
    """
    @staticmethod
    def load(filename, **kwargs) -> LabelBeeRecord:
        json_content = khandy.load_json(filename)
        return LabelBeeRecord(**json_content)

    @staticmethod
    def save(filename, labelbee_record: LabelBeeRecord):
        json_content = dataclasses.asdict(labelbee_record)
        khandy.save_json(filename, json_content, cls=_NumpyEncoder)

    @staticmethod
    def to_ir(labelbee_record: LabelBeeRecord) -> DetectIrRecord:
        ir_record = DetectIrRecord(
            filename='',
            width=labelbee_record.width,
            height=labelbee_record.height
        )
        for labelbee_object in labelbee_record.step_1.result:
            ir_object = DetectIrObject(
                label=labelbee_object.attribute,
                x_min=labelbee_object.x,
                y_min=labelbee_object.y,
                x_max=labelbee_object.x + labelbee_object.width,
                y_max=labelbee_object.y + labelbee_object.height,
            )
            ir_record.objects.append(ir_object)
        return ir_record
        
    @staticmethod
    def from_ir(ir_record: DetectIrRecord) -> LabelBeeRecord:
        labelbee_record = LabelBeeRecord(
            width=ir_record.width,
            height=ir_record.height,
            step_1=LabelBeeStep() # must keep this line
        )
        for k, ir_object in enumerate(ir_record.objects):
            labelbee_object = LabelBeeObject(
                id=f'{ir_object.label}#{k}',
                sourceID='',
                order=k,
                textAttribute='',
                attribute=ir_object.label,
                x=ir_object.x_min,
                y=ir_object.y_min,
                width=ir_object.x_max - ir_object.x_min,
                height=ir_object.y_max - ir_object.y_min
            )
            labelbee_record.step_1.result.append(labelbee_object)
        return labelbee_record


def load_detect(filename, fmt, **kwargs) -> Union[DetectIrRecord, List[DetectIrRecord]]:
    if fmt == 'labelme':
        labelme_record = LabelmeHandler.load(filename, **kwargs)
        ir_record = LabelmeHandler.to_ir(labelme_record)
    elif fmt == 'yolo':
        yolo_record = YoloHandler.load(filename, **kwargs)
        ir_record = YoloHandler.to_ir(yolo_record)
    elif fmt in ('voc', 'pascal', 'pascal_voc'):
        pascal_voc_record = PascalVocHandler.load(filename, **kwargs)
        ir_record = PascalVocHandler.to_ir(pascal_voc_record)
    elif fmt == 'coco':
        coco_records = CocoHandler.load(filename, **kwargs)
        ir_record = [CocoHandler.to_ir(coco_record) for coco_record in coco_records]
    elif fmt == 'labelbee':
        labelbee_record = LabelBeeHandler.load(filename, **kwargs)
        ir_record = LabelBeeHandler.to_ir(labelbee_record)
    else:
        raise ValueError(f"Unsupported detect label fmt. Got {fmt}")
    return ir_record


def save_detect(filename, ir_record: DetectIrRecord, out_fmt):
    os.makedirs(os.path.dirname(os.path.abspath(filename)), exist_ok=True)
    if out_fmt == 'labelme':
        labelme_record = LabelmeHandler.from_ir(ir_record)
        LabelmeHandler.save(filename, labelme_record)
    elif out_fmt == 'yolo':
        yolo_record = YoloHandler.from_ir(ir_record)
        YoloHandler.save(filename, yolo_record)
    elif out_fmt in ('voc', 'pascal', 'pascal_voc'):
        pascal_voc_record = PascalVocHandler.from_ir(ir_record)
        PascalVocHandler.save(filename, pascal_voc_record)
    elif out_fmt == 'coco':
        raise ValueError("Unsupported for `coco` now!")
    elif out_fmt == 'labelbee':
        labelbee_record = LabelBeeHandler.from_ir(ir_record)
        LabelBeeHandler.save(filename, labelbee_record)
    else:
        raise ValueError(f"Unsupported detect label fmt. Got {out_fmt}")


def _get_format(record):
    if isinstance(record, LabelmeRecord):
        return ('labelme',)
    elif isinstance(record, YoloRecord):
        return ('yolo',)
    elif isinstance(record, PascalVocRecord):
        return ('voc', 'pascal', 'pascal_voc')
    elif isinstance(record, CocoRecord):
        return ('coco',)
    elif isinstance(record, LabelBeeRecord):
        return ('labelbee',)
    elif isinstance(record, DetectIrRecord):
        return ('ir', 'detect_ir')
    else:
        return ()


def convert_detect(record, out_fmt):
    allowed_fmts = ('labelme', 'yolo', 'voc', 'coco', 'pascal', 'pascal_voc', 
                    'labelbee', 'ir', 'detect_ir')
    if out_fmt not in allowed_fmts:
        raise ValueError("Unsupported label format conversions for given out_fmt")
    if out_fmt in _get_format(record):
        return record

    if isinstance(record, LabelmeRecord):
        ir_record = LabelmeHandler.to_ir(record)
    elif isinstance(record, YoloRecord):
        ir_record = YoloHandler.to_ir(record)
    elif isinstance(record, PascalVocRecord):
        ir_record = PascalVocHandler.to_ir(record)
    elif isinstance(record, CocoRecord):
        ir_record = CocoHandler.to_ir(record)
    elif isinstance(record, LabelBeeRecord):
        ir_record = LabelBeeHandler.to_ir(record)
    elif isinstance(record, DetectIrRecord):
        ir_record = record
    else:
        raise TypeError('Unsupported type for record')

    if out_fmt in ('ir', 'detect_ir'):
        dst_record = ir_record
    elif out_fmt == 'labelme':
        dst_record = LabelmeHandler.from_ir(ir_record)
    elif out_fmt == 'yolo':
        dst_record = YoloHandler.from_ir(ir_record)
    elif out_fmt in ('voc', 'pascal', 'pascal_voc'):
        dst_record = PascalVocHandler.from_ir(ir_record)
    elif out_fmt == 'coco':
        dst_record = CocoHandler.from_ir(ir_record)
    elif out_fmt == 'labelbee':
        dst_record = LabelBeeHandler.from_ir(ir_record)
    return dst_record


def replace_detect_label(record: DetectIrRecord, label_map: Dict[str, str], ignore: bool = True) -> DetectIrRecord:
    dst_record = copy.deepcopy(record)
    dst_objects = []
    for ir_object in dst_record.objects:
        if not ignore:
            if ir_object.label in label_map:
                ir_object.label = label_map[ir_object.label]
            dst_objects.append(ir_object)
        else:
            if ir_object.label in label_map:
                ir_object.label = label_map[ir_object.label]
                dst_objects.append(ir_object)
    dst_record.objects = dst_objects
    return dst_record


def load_coco_class_names(filename):
    warnings.warn('load_coco_class_names will be deprecated, use CocoHandler.load_class_names instead!')
    json_data = khandy.load_json(filename)
    categories = json_data['categories']
    return [cat_item['name'] for cat_item in categories]


def crop_detect(ir_record: DetectIrRecord, x_min, y_min, x_max, y_max, min_area_ratio) -> DetectIrRecord:
    """Crops DetectIrRecord based on a bounding box and a minimum area ratio.

    Args:
        ir_record ( DetectIrRecord): The DetectIrRecord to process.
        x_min (int): The minimum x-coordinate of the cropping box.
        y_min (int): The minimum y-coordinate of the cropping box.
        x_max (int): The maximum x-coordinate of the cropping box.
        y_max (int): The maximum y-coordinate of the cropping box.
        min_area_ratio (float): The minimum intersection over area (IOA) ratio for an object to be considered within the cropped box.

    Returns:
        DetectIrRecord: A new DetectIrRecord containing the cropped objects.

    Raises:
        AssertionError: If any of the coordinate values are not integers or if the cropping box is not valid (x_min > x_max or y_min > y_max).

    See Also:
        crop_image
    """
    assert isinstance(x_min, numbers.Integral) and isinstance(y_min, numbers.Integral)
    assert isinstance(x_max, numbers.Integral) and isinstance(y_max, numbers.Integral)
    assert (x_min < x_max) and (y_min < y_max)

    dst_ir_record = DetectIrRecord('', width=x_max - x_min, height=y_max - y_min)
    cropped_box = [x_min, y_min, x_max, y_max]
    cropped_box = np.asarray(cropped_box, dtype=np.float32).reshape(1, -1)
    for ir_object in ir_record.objects:
        object_box = [ir_object.x_min, ir_object.y_min, ir_object.x_max, ir_object.y_max]
        object_box = np.asarray(object_box, dtype=np.float32).reshape(1, -1)
        ioa = khandy.paired_overlap_ratio(cropped_box, object_box, 'ioa').item()
        if ioa >= min_area_ratio:
            max_x_mins = np.maximum(cropped_box[:, 0], object_box[:, 0])
            max_y_mins = np.maximum(cropped_box[:, 1], object_box[:, 1])
            min_x_maxs = np.minimum(cropped_box[:, 2], object_box[:, 2])
            min_y_maxs = np.minimum(cropped_box[:, 3], object_box[:, 3])
            new_ir_object = DetectIrObject(
                label=ir_object.label,
                x_min=max_x_mins[0] - x_min,
                y_min=max_y_mins[0] - y_min,
                x_max=min_x_maxs[0] - x_min,
                y_max=min_y_maxs[0] - y_min)
            dst_ir_record.objects.append(new_ir_object)
    return dst_ir_record


def fliter_detect(
    detect_ir: Union[DetectIrRecord, List[DetectIrObject]], 
    func: Callable[[DetectIrObject], bool]
) -> Union[DetectIrRecord, List[DetectIrObject]]:
    if isinstance(detect_ir, DetectIrRecord):
        return DetectIrRecord(
            filename=detect_ir.filename,
            width=detect_ir.width,
            height=detect_ir.height,
            objects=list(filter(func, detect_ir.objects))
        )
    elif khandy.is_list_of(detect_ir, DetectIrObject):
        return list(filter(func, detect_ir))
    else:
        raise TypeError('detect_ir must be DetectIrRecord or list of DetectIrObject')
    
    
def fliter_detect_by_label(
    detect_ir: Union[DetectIrRecord, List[DetectIrObject]], 
    labels: Union[str, List[str], Callable[[str], bool]]
) -> Union[DetectIrRecord, List[DetectIrObject]]:
    if not callable(labels):
        labels = khandy.to_list(labels)
        label_checker = lambda x: x in labels
    else:
        label_checker = labels
    return fliter_detect(detect_ir, lambda obj: label_checker(obj.label))




from collections import OrderedDict

import khandy
import numpy as np


def convert_feature_dict_to_array(feature_dict):
    one_feature = khandy.get_dict_first_item(feature_dict)[1]
    num_features = sum([len(item) for item in feature_dict.values()])
    
    key_list = []
    start_index = 0
    feature_array = np.empty((num_features, one_feature.shape[-1]), one_feature.dtype)
    for key, value in feature_dict.items():
        feature_array[start_index: start_index + len(value)]= value
        key_list += [key] * len(value)
        start_index += len(value)
    return key_list, feature_array


def convert_feature_array_to_dict(key_list, feature_array):
    assert len(key_list) == len(feature_array)
    feature_dict = OrderedDict()
    for key, feat in zip(key_list, feature_array):
        feature_dict.setdefault(key, []).append(feat)
    for label in feature_dict.keys():
        feature_dict[label] = np.vstack(feature_dict[label])
    return feature_dict
    
    
def pairwise_distances(x, y, squared=True):
    """Compute pairwise (squared) Euclidean distances.
    
    References:
        [2016 CVPR] Deep Metric Learning via Lifted Structured Feature Embedding
        `euclidean_distances` from sklearn
    """
    assert isinstance(x, np.ndarray) and x.ndim == 2
    assert isinstance(y, np.ndarray) and y.ndim == 2
    assert x.shape[1] == y.shape[1]
    
    x_square = np.expand_dims(np.einsum('ij,ij->i', x, x), axis=1)
    if x is y:
        y_square = x_square.T
    else:
        y_square = np.expand_dims(np.einsum('ij,ij->i', y, y), axis=0)
    distances = np.dot(x, y.T)
    # use inplace operation to accelerate
    distances *= -2
    distances += x_square
    distances += y_square
    # result maybe less than 0 due to floating point rounding errors.
    np.maximum(distances, 0, distances)
    if x is y:
        # Ensure that distances between vectors and themselves are set to 0.0.
        # This may not be the case due to floating point rounding errors.
        distances.flat[::distances.shape[0] + 1] = 0.0
    if not squared:
        np.sqrt(distances, distances)
    return distances
    
    


import numpy as np


def _concat(arr_list, axis=0):
    """Avoids a copy if there is only a single element in a list.
    """
    if len(arr_list) == 1:
        return arr_list[0]
    return np.concatenate(arr_list, axis)
    
    
def convert_boxes_list_to_boxes_and_indices(boxes_list):
    """
    Args:
        boxes_list (np.ndarray): list or tuple of ndarray with shape (N_i, 4+K)
        
    Returns:
        boxes (ndarray): shape (M, 4+K) where M is sum of N_i.
        indices (ndarray): shape (M, 1) where M is sum of N_i.
        
    References:
        `mmdet.core.bbox.bbox2roi` in mmdetection
        `convert_boxes_to_roi_format` in TorchVision
        `modeling.poolers.convert_boxes_to_pooler_format` in detectron2
    """
    assert isinstance(boxes_list, (list, tuple))
    boxes = _concat(boxes_list, axis=0)
    
    indices_list = [np.full((len(b), 1), i, boxes.dtype) 
                    for i, b in enumerate(boxes_list)]
    indices = _concat(indices_list, axis=0)
    return boxes, indices
    
    
def convert_boxes_and_indices_to_boxes_list(boxes, indices, num_indices):
    """
    Args:
        boxes (np.ndarray): shape (N, 4+K)
        indices (np.ndarray): shape (N,) or (N, 1), maybe batch index 
            in mini-batch or class label index.
        num_indices (int): number of index.

    Returns:
        list (ndarray): boxes list of each index
        
    References:
        `mmdet.core.bbox2result` in mmdetection
        `mmdet.core.bbox.roi2bbox` in mmdetection
        `convert_boxes_to_roi_format` in TorchVision
        `modeling.poolers.convert_boxes_to_pooler_format` in detectron2
    """
    boxes = np.asarray(boxes)
    indices = np.asarray(indices)
    assert boxes.ndim == 2, "boxes ndim must be 2, got {}".format(boxes.ndim)
    assert (indices.ndim == 1) or (indices.ndim == 2 and indices.shape[-1] == 1), \
        "indices ndim must be 1 or 2 if last dimension size is 1, got shape {}".format(indices.shape)
    assert boxes.shape[0] == indices.shape[0], "the 1st dimension size of boxes and indices "\
        "must be the same, got {} != {}".format(boxes.shape[0], indices.shape[0])
        
    if boxes.shape[0] == 0:
        return [np.zeros((0, boxes.shape[1]), dtype=np.float32) 
                for i in range(num_indices)]
    else:
        if indices.ndim == 2:
            indices = np.squeeze(indices, axis=-1)
        return [boxes[indices == i, :] for i in range(num_indices)]
    
    


import collections.abc
import itertools
import numbers
import random
import warnings
from typing import Any, List, Optional, Union

import numpy as np


def to_list(obj) -> Optional[List]:
    if obj is None:
        return None
    elif hasattr(obj, '__iter__') and not isinstance(obj, str):
        try:
            return list(obj)
        except:
            return [obj]
    else:
        return [obj]


def convert_lists_to_record(*list_objs, delimiter=None):
    warnings.warn('convert_lists_to_record will be deprecated, use convert_table_to_records or/with transpose_table instead!')

    assert len(list_objs) >= 1, 'list_objs length must >= 1.'
    delimiter = delimiter or ','

    assert isinstance(list_objs[0], (tuple, list))
    number = len(list_objs[0])
    for item in list_objs[1:]:
        assert isinstance(item, (tuple, list))
        assert len(item) == number, '{} != {}'.format(len(item), number)

    records = []
    record_list = zip(*list_objs)
    for record in record_list:
        record_str = [str(item) for item in record]
        records.append(delimiter.join(record_str))
    return records


def convert_table_to_records(*table, delimiter=None):
    """Convert a table of tuples or lists into a list of records separated by a delimiter.

    Args:
        *table: A variable number of tuples or lists representing rows in the table.
        delimiter (str, optional): The delimiter to use for separating values in each record. Defaults to ','.

    Returns:
        list: A list of strings, where each string is a record representing a row in the table.

    Raises:
        AssertionError: If the length of 'table' is less than 1.
        AssertionError: If the first item in 'table' is not a tuple or list.
        AssertionError: If any subsequent item in 'table' is not a tuple or list of the same length as the first item.
    """
    assert len(table) >= 1, 'table length must >= 1.'
    delimiter = delimiter or ','

    assert isinstance(table[0], (tuple, list))
    number = len(table[0])
    for item in table[1:]:
        assert isinstance(item, (tuple, list))
        assert len(item) == number, '{} != {}'.format(len(item), number)

    records = [delimiter.join(str(item) for item in row) for row in table]
    return records


def shuffle_table(table):
    """
    Notes:
        table can be seen as sequence of sequence which have equal items.
    """
    shuffled = list(zip(*table))
    random.shuffle(shuffled)
    dst_table = tuple(zip(*shuffled))
    return dst_table


def transpose_table(table):
    """
    Notes:
        table can be seen as sequence of sequence which have equal items.
    """
    return tuple(zip(*table))


def concat_list(in_list):
    """Concatenate a list of list into a single list.

    Args:
        in_list (list): The list of list to be merged.

    Returns:
        list: The concatenated flat list.

    References:
        mmcv.concat_list
    """
    return list(itertools.chain(*in_list))


def split_by_num(x, num_splits, strict=True):
    """
    Args:
        num_splits: an integer indicating the number of splits

    References:
        numpy.split and numpy.array_split
    """
    # NB: np.ndarray is not Sequence
    assert isinstance(x, (collections.abc.Sequence, np.ndarray))
    assert isinstance(num_splits, numbers.Integral)

    if strict:
        assert len(x) % num_splits == 0
    split_size = (len(x) + num_splits - 1) // num_splits
    out_list = []
    for i in range(0, len(x), split_size):
        out_list.append(x[i: i + split_size])
    return out_list


def split_by_size(x, sizes):
    """
    References:
        tf.split
        https://github.com/open-mmlab/mmcv/blob/master/mmcv/utils/misc.py
    """
    # NB: np.ndarray is not Sequence
    assert isinstance(x, (collections.abc.Sequence, np.ndarray))
    assert isinstance(sizes, (list, tuple))

    assert sum(sizes) == len(x)
    out_list = []
    start_index = 0
    for size in sizes:
        out_list.append(x[start_index: start_index + size])
        start_index += size
    return out_list


def split_by_slice(x, slices):
    """
    References:
        SliceLayer in Caffe, and numpy.split
    """
    # NB: np.ndarray is not Sequence
    assert isinstance(x, (collections.abc.Sequence, np.ndarray))
    assert isinstance(slices, (list, tuple))

    out_list = []
    indices = [0] + list(slices) + [len(x)]
    for i in range(len(slices) + 1):
        out_list.append(x[indices[i]: indices[i + 1]])
    return out_list


def split_by_ratio(x, ratios):
    # NB: np.ndarray is not Sequence
    assert isinstance(x, (collections.abc.Sequence, np.ndarray))
    assert isinstance(ratios, (list, tuple))

    pdf = [k / sum(ratios) for k in ratios]
    cdf = [sum(pdf[:k]) for k in range(len(pdf) + 1)]
    indices = [int(round(len(x) * k)) for k in cdf]
    return [x[indices[i]: indices[i + 1]] for i in range(len(ratios))]


def to_ntuple(x, n):
    """Convert the input into a tuple of length n.

    Args:
        x (Any): The input to be converted.
        n (numbers.Integral): The length of the resulting tuple, which must be a positive integer.

    Returns:
        tuple: A tuple of length n.

    Raises:
        ValueError: If n is not a positive integer.
        AssertionError: If len(x) is not equal to n when x is a non-str sequence.
    """
    if not (isinstance(n, numbers.Integral) and n > 0):
        raise ValueError(f'n must be positive integer, got {n}.')
    if isinstance(x, collections.abc.Sequence) and not isinstance(x, str):
        assert len(x) == n, f'len(x) is not equal to {n}.'
        return tuple(x)
    return tuple(x for _ in range(n))


def to_1tuple(x):
    return to_ntuple(x, 1)


def to_2tuple(x):
    return to_ntuple(x, 2)


def to_3tuple(x):
    return to_ntuple(x, 3)


def to_4tuple(x):
    return to_ntuple(x, 4)


def is_seq_of(seq, item_type, seq_type=None) -> bool:
    """Check whether it is a sequence of some type.

    Args:
        seq (Sequence): The sequence to be checked.
        item_type (type): Expected type of sequence items.
        seq_type (type, optional): Expected sequence type.

    Returns:
        bool: Whether the sequence is valid.

    References:
        mmcv
    """
    if seq_type is None:
        exp_seq_type = collections.abc.Sequence
    else:
        assert issubclass(seq_type, collections.abc.Sequence)
        exp_seq_type = seq_type
    if not isinstance(seq, exp_seq_type):
        return False
    for item in seq:
        if not isinstance(item, item_type):
            return False
    return True


def is_list_of(seq, item_type):
    """Check whether it is a list of some type.

    A partial method of :func:`is_seq_of`.
    """
    return is_seq_of(seq, item_type, seq_type=list)


def is_tuple_of(seq, item_type):
    """Check whether it is a tuple of some type.

    A partial method of :func:`is_seq_of`.
    """
    return is_seq_of(seq, item_type, seq_type=tuple)


class EqLenSequences:
    """Class for managing sequences of equal length.

    References:
        https://github.com/facebookresearch/detectron2/blob/main/detectron2/structures/instances.py
        https://github.com/open-mmlab/mmdetection/blob/master/mmdet/core/data_structures/instance_data.py
    """
    def __init__(self, **kwargs) -> None:
        """Initialize the class object with keyword arguments.

        Args:
            **kwargs: Keyword arguments where the key is the attribute name and the value is the attribute value.
        """
        self._fields = set()
        for k, v in kwargs.items():
            setattr(self, k, v)

    def __setattr__(self, name: str, value: Any) -> None:
        """Override the default attribute assignment.
        Only allows assignment of sequences (objects with __len__ and __getitem__ attributes).

        Args:
            name (str): Name of the attribute to set.
            value (Any): Value of the attribute, which should be a sequence.

        Raises:
            AttributeError: If trying to change the private attribute '_fields'.
            AssertionError: If the value is not a sequence or its length is inconsistent with existing sequences.
        """
        if name == '_fields':
            # hasattr is implemented by calling getattr(object, name) and
            # seeing whether it raises an AttributeError or not.
            if not hasattr(self, name):
                super().__setattr__(name, value)
            else:
                raise AttributeError(f'{name} has been used as a private attribute, which is immutable.')
        else:
            # colllections.abc.Sequence is not used here to avoid issues with
            # np.ndarray and torch.Tensor which is not a Sequence
            assert hasattr(value, '__len__'), 'value must contain `__len__` attribute'
            assert hasattr(value, '__getitem__'), 'value must contain `__getitem__` attribute'
            if len(self._fields) > 0:
                msg = f'{name} len is not consistent with len(self), {len(value)} vs {len(self)}'
                assert len(value) == len(self), msg
            self._fields.add(name)
            super().__setattr__(name, value)

    def __delattr__(self, name: str) -> None:
        """Override the default attribute deletion.

        Args:
            name (str): Name of the attribute to delete.

        Raises:
            AttributeError: If trying to delete the private attribute '_fields'.
        """
        if name == '_fields':
            raise AttributeError(f'{name} has been used as a private attribute, which is immutable.')
        super().__delattr__(name)
        self._fields.remove(name)

    def __len__(self) -> int:
        """Returns the length of the sequences stored in the object.

        Returns:
            int: The length of the sequences, or 0 if there are no fields.

        Note:
            All sequences must have the same length.
        """
        for name in self._fields:
            return len(getattr(self, name))
        return 0

    def __contains__(self, name: str) -> bool:
        """Checks if the given name is a field (sequence) in the object.

        Args:
            name (str): Name of the field to check.

        Returns:
            bool: True if the field exists, False otherwise.
        """
        return name in self._fields

    def __getitem__(self, key: Union[int, slice, List[int], np.ndarray]):
        """Supports indexing and slicing to get a new object
        with sequences indexed or sliced by the given key.

        Args:
            key (Union[int, slice, List[int]], np.ndarray): Index, slice, or list of indices to use for getting sequences.

        Returns:
            A new object of the same class with sequences indexed or sliced according to the key.

        Raises:
            IndexError: If the index, slice, or any index in the list is out of range.
            TypeError: If the key is not an integer, slice, list of integers, or a 1D numpy array.
            AssertionError: If the key is a numpy array but not 1D, or if the length of the sequence 
                does not match the length of the key when using a boolean mask.
        """
        if type(key) is int:
            if key >= len(self) or key < -len(self):
                raise IndexError(f"{self.__class__.__name__} index {key} out of range!")
            else:
                key = slice(key, None, len(self))
        elif isinstance(key, np.ndarray):
            assert key.ndim == 1, f"Key must be a 1D array, but got {key.ndim}D array."
        elif not isinstance(key, (slice, list)):
            raise TypeError(f"Unsupported key type: {type(key)}")
            
        kwargs = {}
        for name in self._fields:
            value = getattr(self, name)
            try:
                kwargs[name] = value[key]
            except:
                if isinstance(key, np.ndarray) and key.dtype == np.bool_:
                    assert len(value) == len(key), \
                        f"Length of {name} ({len(value)}) does not match length of key ({len(key)})"
                    # If key is a boolean array, use it to index the sequence
                    kwargs[name] = [value[i] for i, elem in enumerate(key) if elem]
                else:
                    kwargs[name] = [value[i] for i in key]
        return self.__class__(**kwargs)

    def __str__(self) -> str:
        """Return a string representation of the EqLenSequences object.

        Returns:
            str: A string representation of the object.
        """
        s = self.__class__.__name__ + "("
        s += "num_instances={}, ".format(len(self))
        s += "fields={{{}}})".format(", ".join((f"{name}: {getattr(self, name)}" for name in self._fields)))
        return s

    __repr__ = __str__

    def get_fields(self) -> set:
        """Return a set of attribute names (fields) stored in the object.

        Returns:
            set: A set of attribute names.
        """
        return self._fields

    def filter_(self, index: Union[int, slice, List[int], np.ndarray]):
        """Filter the EqLenSequences object by indexing the stored sequences using the provided index.

        This method allows you to filter the sequences stored in the EqLenSequences object by
        specifying an integer, slice, or list of integers as the index. Depending on the type of
        index provided, the corresponding elements or slices will be retained in the object.

        Args:
            index (Union[int, slice, List[int], np.ndarray]): The index or slice to use for filtering.
                - If an integer is provided, it should be within the valid range of the object's length.
                - If a slice is provided, it will be used to slice the sequences.
                - If a list of integers is provided, the sequences at the specified indices will be retained.
                - If a numpy array is provided, it should be a 1D array of integers or booleans:
                    - If a boolean numpy array is provided, it will be used to filter the sequences based on the mask.
                    - If a list of integers is provided, it can be used to index the sequences.

        Returns:
            EqLenSequences: The filtered EqLenSequences object with the selected sequences.

        Raises:
            IndexError: If the provided index is out of range (i.e., an integer index is greater than
                or equal to the length of the object, or less than negative the length of the object).
            TypeError: If the provided index is not an integer, slice, list of integers, or a 1D numpy array.
            AssertionError: If the provided index is a numpy array but not 1D, or if the length of the sequence
                does not match the length of the key when using a boolean mask.
        """
        if type(index) is int:
            if index >= len(self) or index < -len(self):
                raise IndexError(f"{self.__class__.__name__} index {index} out of range!")
            else:
                index = slice(index, None, len(self))
        elif isinstance(index, np.ndarray):
            assert index.ndim == 1, f"Index must be a 1D array, but got {index.ndim}D array."
        elif not isinstance(index, (slice, list)):
            raise TypeError(f"Unsupported index type: {type(index)}")

        for name in self._fields:
            value = getattr(self, name)
            try:
                super().__setattr__(name, value[index])
            except:
                # NB: used to avoid the following errors:
                # TypeError: only integer scalar arrays can be converted to a scalar index
                # TypeError: list indices must be integers or slices, not list
                if isinstance(index, np.ndarray) and index.dtype == np.bool_:
                    assert len(value) == len(index), \
                        f"Length of {name} ({len(value)}) does not match length of key ({len(index)})"
                    # If key is a boolean array, use it to index the sequence
                    super().__setattr__(name, [value[i] for i, elem in enumerate(index) if elem])
                else:
                    super().__setattr__(name, [value[i] for i in index])
        return self
    
    def filter(self, index: Union[int, slice, List[int], np.ndarray], inplace: bool):
        if inplace:
            return self.filter_(index)
        else:
            return self[index]
        


import numbers

import khandy


def translate_image(image, x_shift, y_shift, border_value=0):
    """Translate an image.
    
    Args:
        image (ndarray): Image to be translated with format (h, w) or (h, w, c).
        x_shift (int): The offset used for translate in horizontal
            direction. right is the positive direction.
        y_shift (int): The offset used for translate in vertical
            direction. down is the positive direction.
        border_value (int | tuple[int]): Value used in case of a 
            constant border.
            
    Returns:
        ndarray: The translated image.

    See Also:
        crop_image
    """
    assert khandy.is_numpy_image(image)
    assert isinstance(x_shift, numbers.Integral)
    assert isinstance(y_shift, numbers.Integral)
    image_height, image_width = image.shape[:2]
    channels = 1 if image.ndim == 2 else image.shape[2]
    
    if isinstance(border_value, (tuple, list)):
        assert len(border_value) == channels, \
            'Expected the num of elements in tuple equals the channels ' \
            'of input image. Found {} vs {}'.format(
                len(border_value), channels)
    else:
        border_value = (border_value,) * channels
    dst_image = khandy.create_solid_color_image(
        image_height, image_width, border_value, dtype=image.dtype)
    
    if (abs(x_shift) >= image_width) or (abs(y_shift) >= image_height):
        return dst_image
        
    src_x_begin = max(-x_shift, 0)
    src_x_end   = min(image_width - x_shift, image_width)
    dst_x_begin = max(x_shift, 0)
    dst_x_end   = min(image_width + x_shift, image_width)
    
    src_y_begin = max(-y_shift, 0)
    src_y_end   = min(image_height - y_shift, image_height)
    dst_y_begin = max(y_shift, 0)
    dst_y_end   = min(image_height + y_shift, image_height)
    
    dst_image[dst_y_begin:dst_y_end, dst_x_begin:dst_x_end] = \
        image[src_y_begin:src_y_end, src_x_begin:src_x_end]
    return dst_image
    
    


from abc import ABC, abstractmethod

import numpy as np

import khandy

__all__ = ['Extractor']


class Extractor(ABC):
    def __init__(self, feature_dim: int):
        self._feature_dim = feature_dim

    @property
    def feature_dim(self) -> int:
        return self._feature_dim
    
    @abstractmethod
    def forward(self, image: khandy.KArray, **kwargs) -> khandy.KArray:
        pass

    def __call__(self, image: khandy.KArray, **kwargs) -> khandy.KArray:
        return self.forward(image, **kwargs)
    
    





from typing import Optional, Union
import numpy as np


def find_max_area_box(boxes):
    """Find the index of the bounding box with the maximum area.

    Args:
        boxes (numpy array): A 2D numpy array of shape (N, 4) representing a group of bounding boxes. 
            Each row in the array represents a bounding box with the coordinates of the top-left and 
            bottom-right corners.
            
    Returns:
        int: An integer value representing the index of the bounding box with the maximum area in 
            the input 'boxes' array.
    """
    widths = boxes[:, 2] - boxes[:, 0]
    heights = boxes[:, 3] - boxes[:, 1]
    areas = widths * heights
    return np.argmax(areas)


def sort_boxes_by_area(boxes, reverse=False):
    """Sort bounding boxes based on their areas.

    Args:
        boxes (numpy array): A 2D numpy array of shape (N, 4) representing a group of bounding boxes. 
            Each row in the array represents a bounding box with the coordinates of the top-left and 
            bottom-right corners.
            
        reverse (bool): A boolean value indicating whether to sort the bounding boxes in descending order 
            instead of ascending order of their areas. Defaults to False.
    
    Returns:
        int array: A numpy array of integers representing the indices of the bounding boxes sorted 
            by their areas in the input 'boxes' array. The indices can be used to sort the bounding boxes 
            based on their areas. If 'reverse' is set to True, the bounding boxes will be sorted in 
            descending order of their areas.
    """
    widths = boxes[:, 2] - boxes[:, 0]
    heights = boxes[:, 3] - boxes[:, 1]
    areas = widths * heights
    if not reverse:
        return np.argsort(areas)
    else:
        return np.argsort(areas)[::-1]
    

def filter_small_boxes(
    boxes: np.ndarray, 
    min_width: Optional[Union[int, float]] = None, 
    min_height: Optional[Union[int, float]] = None
) -> np.ndarray:
    """Filter bounding boxes based on minimum width and height requirements.

    Args:
        boxes (numpy array): A 2D numpy array of shape (N, 4) representing a group of bounding boxes. 
            Each row in the array represents a bounding box with the coordinates of the top-left and 
            bottom-right corners.
        min_width (int, float or None): The minimum allowed width of the bounding boxes. If set to None, 
            no width requirement will be enforced.
        min_height (int, float or None): The minimum allowed height of the bounding boxes. If set to None, 
            no height requirement will be enforced.
    
    Returns:
        int array: A numpy array of integers representing the indices of the bounding boxes in the 
            input 'boxes' array that meet the minimum width and height requirements. The indices can be 
            used to filter the bounding boxes based on their dimensions.
        
    Raises:
        ValueError: If both min_width and min_height are set to None at the same time.
        
    References:
        `_filter_boxes` in py-faster-rcnn
        `prune_small_boxes` in TensorFlow object detection API.
        `structures.Boxes.nonempty` in detectron2
        `ops.boxes.remove_small_boxes` in torchvision
    """
    mask = np.ones((len(boxes),), dtype=bool)
    if min_width is not None:
        widths = boxes[:, 2] - boxes[:, 0]
        mask &= (widths >= min_width)
    if min_height is not None:
        heights = boxes[:, 3] - boxes[:, 1]
        mask &= (heights >= min_height)
    return np.nonzero(mask)[0]
    

def filter_boxes_outside(boxes, reference_box):
    """Filters bounding boxes that fall outside reference box.
    
    References:
        `prune_outside_window` in TensorFlow object detection API.
    """
    x_min, y_min, x_max, y_max = reference_box[:4]
    mask = ((boxes[:, 0] >= x_min) & (boxes[:, 1] >= y_min) &
            (boxes[:, 2] <= x_max) & (boxes[:, 3] <= y_max))
    return np.nonzero(mask)[0]


def filter_boxes_completely_outside(boxes, reference_box):
    """Filters bounding boxes that fall completely outside of reference box.
    
    References:
        `prune_completely_outside_window` in TensorFlow object detection API.
    """
    x_min, y_min, x_max, y_max = reference_box[:4]
    mask = ((boxes[:, 0] < x_max) & (boxes[:, 1] < y_max) &
            (boxes[:, 2] > x_min) & (boxes[:, 3] > y_min))
    return np.nonzero(mask)[0]
    

def non_max_suppression(boxes, scores, thresh, classes=None, ratio_type="iou"):
    """Greedily select boxes with high confidence
    Args:
        boxes: [[x_min, y_min, x_max, y_max], ...]
        scores: object confidence
        thresh: retain overlap_ratio <= thresh
        classes: class labels
        
    Returns:
        indices to keep
        
    References:
        `py_cpu_nms` in py-faster-rcnn
        torchvision.ops.nms
        torchvision.ops.batched_nms
    """

    if boxes.size == 0:
        return np.empty((0,), dtype=np.int64)
    if classes is not None:
        # strategy: in order to perform NMS independently per class,
        # we add an offset to all the boxes. The offset is dependent
        # only on the class idx, and is large enough so that boxes
        # from different classes do not overlap
        max_coordinate = np.max(boxes)
        offsets = classes * (max_coordinate + 1)
        if offsets.ndim == 1:
            offsets = offsets[:, None]
        boxes = boxes + offsets

    x_mins = boxes[:, 0]
    y_mins = boxes[:, 1]
    x_maxs = boxes[:, 2]
    y_maxs = boxes[:, 3]
    areas = (x_maxs - x_mins) * (y_maxs - y_mins)
    order = scores.flatten().argsort()[::-1]
    
    keep = []
    while order.size > 0:
        i = order[0]
        keep.append(i)
        
        max_x_mins = np.maximum(x_mins[i], x_mins[order[1:]])
        max_y_mins = np.maximum(y_mins[i], y_mins[order[1:]])
        min_x_maxs = np.minimum(x_maxs[i], x_maxs[order[1:]])
        min_y_maxs = np.minimum(y_maxs[i], y_maxs[order[1:]])
        widths = np.maximum(0, min_x_maxs - max_x_mins)
        heights = np.maximum(0, min_y_maxs - max_y_mins)
        intersect_areas = widths * heights
        
        if ratio_type in ["union", 'iou']:
            ratio = intersect_areas / (areas[i] + areas[order[1:]] - intersect_areas)
        elif ratio_type == "min":
            ratio = intersect_areas / np.minimum(areas[i], areas[order[1:]])
        else:
            raise ValueError('Unsupported ratio_type. Got {}'.format(ratio_type))
            
        inds = np.nonzero(ratio <= thresh)[0]
        order = order[inds + 1]
    return np.asarray(keep)
    


import numpy as np


def assert_and_normalize_shape(x, length):
    """
    Args:
        x: ndarray
        length: int
    """
    if x.ndim == 0:
        return x
    elif x.ndim == 1:
        if len(x) == 1:
            return x
        elif len(x) == length:
            return x
        else:
            raise ValueError('Incompatible shape!')
    elif x.ndim == 2:
        if x.shape == (1, 1):
            return np.squeeze(x, axis=-1)
        elif x.shape == (length, 1):
            return np.squeeze(x, axis=-1)
        else:
            raise ValueError('Incompatible shape!') 
    else:
        raise ValueError('Incompatible ndim!')
        



import unittest
import khandy


class TestDetObjectItem(unittest.TestCase):
    def setUp(self):
        self.item = khandy.model.DetObjectItem(
            x_min=10.0, y_min=20.0, x_max=30.0, y_max=40.0,
            conf=0.95, class_index=1, class_name="person",
            _extra_fields={"colors": ["red"], "sizes": ["large"]}
        )

    def test_field_access(self):
        self.assertEqual(self.item.x_min, 10.0)
        self.assertEqual(self.item.y_min, 20.0)
        self.assertEqual(self.item.x_max, 30.0)
        self.assertEqual(self.item.y_max, 40.0)
        self.assertEqual(self.item.conf, 0.95)
        self.assertEqual(self.item.class_index, 1)
        self.assertEqual(self.item.class_name, "person")
        self.assertEqual(self.item.colors, ["red"])
        self.assertEqual(self.item.sizes, ["large"])
        with self.assertRaises(AttributeError):
            _ = self.item.non_existent_field
            
    def test_post_init(self):
        with self.assertRaises(AssertionError):
            self.item = khandy.model.DetObjectItem(
                x_min=10.0, y_min=20.0, x_max=30.0, y_max=40.0,
                conf=0.95, class_index=1, class_name="person",
                _extra_fields={"class_name": ["red"]}
            )
        with self.assertRaises(AssertionError):
            self.item = khandy.model.DetObjectItem(
                x_min=10.0, y_min=20.0, x_max=30.0, y_max=40.0,
                conf=0.95, class_index=1, class_name="person",
                _extra_fields={"colors": "red"}
            )
        with self.assertRaises(AssertionError):
            self.item = khandy.model.DetObjectItem(
                x_min=10.0, y_min=20.0, x_max=30.0, y_max=40.0,
                conf=0.95, class_index=1, class_name="person",
                _extra_fields={"colors": ["red", 'green']}
            )
            
    def test_field_modification(self):
        self.item.x_min = 10.5
        self.assertAlmostEqual(self.item.x_min, 10.5)
        self.item.colors = ["blue"]
        self.assertEqual(self.item.colors, ["blue"])
        self.item.foobar = ["new_value"]
        self.assertEqual(self.item.foobar, ["new_value"])
        
        with self.assertRaises(AssertionError):
            self.item.colors = "blue"
        with self.assertRaises(AssertionError):
            self.item.colors = ["blue", 'green']
            
    def test_to_det_objects(self):
        det_objects = self.item.to_det_objects()
        self.assertIsInstance(det_objects, khandy.model.DetObjects)
        self.assertEqual(len(det_objects), 1)
        self.assertIsInstance(det_objects[0], khandy.model.DetObjectItem)

    def test_area_property(self):
        self.assertAlmostEqual(self.item.area, 400.0)
        with self.assertRaises(AttributeError):
            # AttributeError: can't set attribute
            self.item.area = 100.0


if __name__ == '__main__':
    unittest.main()
    



import numbers
import warnings
from dataclasses import dataclass
from enum import Enum
from typing import List, Optional, Union

import numpy as np


def is_boxes(arr, on_axis: int = -1, allow_extra: bool = True) -> bool:
    """Check if the array `arr` satisfies the "boxes" condition on a specific axis.
    
    Args:
        arr (np.ndarray): The input array to be checked.
        on_axis (int, optional): The axis on which to check the boxes condition. Defaults to -1, indicating the last axis.
        allow_extra (bool, optional): Whether to allow dimensions greater than 4 on the specified axis. Defaults to True.
        
    Returns:
        bool: True if the dimension on the specified axis is at least 4 (when allow_extra is True) or
            exactly 4 (when allow_extra is False); False otherwise.
  
    Raises:
        np.AxisError: If the `on_axis` is out of bounds for the array `arr`.
        AssertionError: If `on_axis` is not an integer or `allow_extra` is not a boolean.
    """
    assert isinstance(on_axis, numbers.Integral), "on_axis must be an integer."
    assert isinstance(allow_extra, bool), "allow_extra must be an boolean."

    arr = np.array(arr)
    if not (-arr.ndim <= on_axis < arr.ndim):
        raise np.AxisError(on_axis, arr.ndim)
    if allow_extra:
        return arr.shape[on_axis] >= 4
    else:
        return arr.shape[on_axis] == 4
    
    
class BoxFormat(Enum):
    XYXY = 'xyxy'
    XYWH = 'xywh'
    CXCYWH = 'cxcywh'
    

@dataclass
class _BoxFieldInfo:
    name: str
    slices: slice
    included: List[BoxFormat]
    doc: Optional[str] = None
    

class _BoxFieldDescriptor:
    def __init__(self, field_info):
        self.name = field_info.name
        self.slices = field_info.slices
        self.included = field_info.included
        self.__doc__ = field_info.doc

    def _warn_for_include(self, instance, raise_error=False):
        if instance.format not in self.included:
            box_formats = [str(item) for item in self.included]
            if not raise_error:
                warnings.warn(f'to get {self.name}, box format had better to be in {box_formats}, got {instance.format}')
            else:
                raise ValueError(f'to set {self.name}, box format should to be in {box_formats}, got {instance.format}')
                
    def __get__(self, instance, owner):
        # when the attribute is accessed through the owner, instance is None
        if instance is None:
            return self

        self._warn_for_include(instance)
        if instance.format in self.included:
            return instance.get_fields(self.slices)
            
        if instance.format == BoxFormat.XYXY:
            if self.name == 'widths':
                return instance.x_maxs - instance.x_mins
            elif self.name == 'heights':
                return instance.y_maxs - instance.y_mins
            elif self.name == 'x_centers':
                return (instance.x_maxs + instance.x_mins) * 0.5
            elif self.name == 'y_centers':
                return (instance.y_maxs + instance.y_mins) * 0.5
            elif self.name == 'lengths':
                return instance.maxs - instance.mins
            elif self.name == 'centers':
                return (instance.maxs + instance.mins) * 0.5
        elif instance.format == BoxFormat.XYWH:
            if self.name == 'x_maxs':
                return instance.x_mins + instance.widths
            elif self.name == 'y_maxs':
                return instance.y_mins + instance.heights
            elif self.name == 'x_centers':
                return instance.x_mins + instance.widths * 0.5
            elif self.name == 'y_centers':
                return instance.y_mins + instance.heights * 0.5
            elif self.name == 'maxs':
                return instance.mins + instance.lengths
            elif self.name == 'centers':
                return instance.mins + instance.lengths * 0.5
            elif self.name == 'x_coords':
                return np.concatenate([instance.x_mins, instance.x_maxs], instance.on_axis)
            elif self.name == 'y_coords':
                return np.concatenate([instance.y_mins, instance.y_maxs], instance.on_axis)
        elif instance.format == BoxFormat.CXCYWH:
            if self.name == 'x_mins':
                return instance.x_centers - instance.widths * 0.5
            elif self.name == 'y_mins':
                return instance.y_centers - instance.heights * 0.5
            elif self.name == 'x_maxs':
                return instance.x_centers + instance.widths * 0.5
            elif self.name == 'y_maxs':
                return instance.y_centers + instance.heights * 0.5
            elif self.name == 'mins':
                return instance.centers - instance.lengths * 0.5
            elif self.name == 'maxs':
                return instance.centers + instance.lengths * 0.5
            elif self.name == 'x_coords':
                return np.concatenate([instance.x_mins, instance.x_maxs], instance.on_axis)
            elif self.name == 'y_coords':
                return np.concatenate([instance.y_mins, instance.y_maxs], instance.on_axis)

    def __set__(self, instance, value):
        self._warn_for_include(instance, True) # NB: raise_error is True
        instance.set_fields(self.slices, value)


class Boxes:
    """Boxes is a class used to represent bounding boxes. It contains multiple properties and methods 
    that facilitate the manipulation and processing of bounding box data.
    """
    field_infos = {}
    field_infos['x_mins'] = _BoxFieldInfo('x_mins', slice(0, 1), [BoxFormat.XYXY, BoxFormat.XYWH], 'x coords of top left points')
    field_infos['y_mins'] = _BoxFieldInfo('y_mins', slice(1, 2), [BoxFormat.XYXY, BoxFormat.XYWH], 'y coords of top left points')
    field_infos['x_maxs'] = _BoxFieldInfo('x_maxs', slice(2, 3), [BoxFormat.XYXY], 'x coords of bottom right points')
    field_infos['y_maxs'] = _BoxFieldInfo('y_maxs', slice(3, 4), [BoxFormat.XYXY], 'y coords of bottom right points')
    field_infos['x_centers'] = _BoxFieldInfo('x_centers', slice(0, 1), [BoxFormat.CXCYWH], 'x coords of centers')
    field_infos['y_centers'] = _BoxFieldInfo('y_centers', slice(1, 2), [BoxFormat.CXCYWH], 'y coords of centers')
    field_infos['widths'] = _BoxFieldInfo('widths', slice(2, 3), [BoxFormat.CXCYWH, BoxFormat.XYWH], 'widths of boxes')
    field_infos['heights'] = _BoxFieldInfo('heights', slice(3, 4), [BoxFormat.CXCYWH, BoxFormat.XYWH], 'heights of boxes')
    field_infos['mins'] = _BoxFieldInfo('mins', slice(0, 2), [BoxFormat.XYXY, BoxFormat.XYWH], 'top left points')
    field_infos['maxs'] = _BoxFieldInfo('maxs', slice(2, 4), [BoxFormat.XYXY], 'bottom right points')
    field_infos['centers'] = _BoxFieldInfo('centers', slice(0, 2), [BoxFormat.CXCYWH], 'centers of boxes')
    field_infos['lengths'] = _BoxFieldInfo('lengths', slice(2, 4), [BoxFormat.CXCYWH, BoxFormat.XYWH], 'side lengths of boxes')
    field_infos['x_coords'] = _BoxFieldInfo('x_coords', slice(0, None, 2), [BoxFormat.XYXY], 'x coords of boxes')
    field_infos['y_coords'] = _BoxFieldInfo('y_coords', slice(1, None, 2), [BoxFormat.XYXY], 'y coords of boxes')
    field_infos['boxes'] = _BoxFieldInfo('boxes', slice(0, 4), [BoxFormat.XYXY, BoxFormat.XYWH, BoxFormat.CXCYWH], 'boxes')

    x_mins = x1 =_BoxFieldDescriptor(field_infos['x_mins'])
    y_mins = y1 = _BoxFieldDescriptor(field_infos['y_mins'])
    x_maxs = x2 = _BoxFieldDescriptor(field_infos['x_maxs'])
    y_maxs = y2 = _BoxFieldDescriptor(field_infos['y_maxs'])
    x_centers = cx = _BoxFieldDescriptor(field_infos['x_centers'])
    y_centers = cy = _BoxFieldDescriptor(field_infos['y_centers'])
    widths = w = x_lengths = lx = _BoxFieldDescriptor(field_infos['widths'])
    heights = h = y_lengths = ly = _BoxFieldDescriptor(field_infos['heights'])
    mins = p1 = xy_mins = _BoxFieldDescriptor(field_infos['mins'])
    maxs = p2 = xy_maxs = _BoxFieldDescriptor(field_infos['maxs'])
    centers = c = xy_centers = _BoxFieldDescriptor(field_infos['centers'])
    lengths = l = xy_lengths = _BoxFieldDescriptor(field_infos['lengths'])
    x_coords = x = _BoxFieldDescriptor(field_infos['x_coords'])
    y_coords = y = _BoxFieldDescriptor(field_infos['y_coords'])
    boxes = b = _BoxFieldDescriptor(field_infos['boxes'])

    def __init__(self, boxes, on_axis=-1, format: Union[BoxFormat, str] = BoxFormat.XYXY, 
                 dtype=np.float32, allow_extra=True):
        """Initializes a Boxes object.

        Args:
            boxes: A numpy array that stores box information.
            on_axis: An integer indicating which axis to store the box fields. Default is -1 (the last axis). 
            format: A Union type that can be either a BoxFormat enum or a string representing the format of 
                the boxes. Default value is BoxFormat.XYXY.
            dtype: A numpy dtype that specifies the type of the box fields. Default is np.float32.
            allow_extra (bool, optional): Whether to allow dimensions greater than 4 on the specified axis. Defaults to True.
        """
        data = np.asarray(boxes, dtype=dtype)
        assert is_boxes(data, on_axis, allow_extra)

        self.data = data
        self.on_axis = on_axis
        self.format = BoxFormat(format)
        self.allow_extra = allow_extra

    def __repr__(self):
        format_string = f'{self.__class__.__name__}('
        format_string += f'(numpy.ndarray(shape={self.data.shape}), '
        format_string += f'on_axis={self.on_axis}, '
        format_string += f'format={self.format}, '
        format_string += f'dtype={self.dtype}, '
        format_string += f'allow_extra={self.allow_extra})'
        return format_string
        
    @property
    def count(self):
        """Returns the number of bounding boxes in the data array.
        """
        return self.data.size // self.data.shape[self.on_axis]
        
    @property
    def shape(self):
        """Returns the shape of the data array.
        """
        return self.data.shape
        
    @property
    def dtype(self):
        """Returns the data type of the bounding box coordinates.
        """
        return self.data.dtype
        
    @property
    def areas(self):
        """Calculates and returns the areas of all bounding boxes in the data array.
        """
        return self.widths * self.heights
        
    def copy(self):
        """Returns a copy of the Boxes object with a deep copy of the data array.
        """
        return Boxes(self.data.copy(), on_axis=self.on_axis, format=self.format,
                     dtype=self.dtype, allow_extra=self.allow_extra)
        
    def get_fields(self, indices):
        """Returns specific fields from the bounding boxes based on the provided indices.
        """
        data_slice = [slice(None)] * self.data.ndim
        data_slice[self.on_axis] = indices
        return self.data[tuple(data_slice)] 

    def set_fields(self, indices, value):
        """Set specific fields of the bounding boxes based on the provided indices and value.
        """
        data_slice = [slice(None)] * self.data.ndim
        data_slice[self.on_axis] = indices
        self.data[tuple(data_slice)] = value
        
    def to(self, new_format, copy=True):
        """Converts the bounding box format to the specified new format. 
        Optionally returns a deep copy of the data array if copy is True (default).
        """
        boxes = self.copy() if copy else self
        new_format = BoxFormat(new_format)
        if boxes.format == new_format:
            return boxes

        if (boxes.format, new_format) == (BoxFormat.XYXY, BoxFormat.XYWH):
            boxes.x_maxs -= boxes.x_mins
            boxes.y_maxs -= boxes.y_mins
            boxes.format = BoxFormat.XYWH
        elif (boxes.format, new_format) == (BoxFormat.XYWH, BoxFormat.XYXY):
            boxes.widths += boxes.x_mins
            boxes.heights += boxes.y_mins
            boxes.format = BoxFormat.XYXY
        elif (boxes.format, new_format) == (BoxFormat.XYWH, BoxFormat.CXCYWH):
            boxes.x_mins += boxes.widths * 0.5
            boxes.y_mins += boxes.heights * 0.5
            boxes.format = BoxFormat.CXCYWH
        elif (boxes.format, new_format) == (BoxFormat.CXCYWH, BoxFormat.XYWH):
            boxes.x_centers -= boxes.widths * 0.5
            boxes.y_centers -= boxes.heights * 0.5
            boxes.format = BoxFormat.XYWH
        elif (boxes.format, new_format) == (BoxFormat.XYXY, BoxFormat.CXCYWH):
            boxes.x_maxs -= boxes.x_mins
            boxes.y_maxs -= boxes.y_mins
            boxes.x_mins += boxes.x_maxs * 0.5
            boxes.y_mins += boxes.y_maxs * 0.5
            boxes.format = BoxFormat.CXCYWH
        elif (boxes.format, new_format) == (BoxFormat.CXCYWH, BoxFormat.XYXY):
            boxes.x_centers -= boxes.widths * 0.5
            boxes.y_centers -= boxes.heights * 0.5
            boxes.widths += boxes.x_centers
            boxes.heights += boxes.y_centers
            boxes.format = BoxFormat.XYXY
        return boxes
    



import hashlib


def calc_hash(content, hash_object=None):
    hash_object = hash_object or hashlib.md5()
    if isinstance(hash_object, str):
        hash_object = hashlib.new(hash_object)
    hash_object.update(content)
    return hash_object.hexdigest()


def calc_file_hash(filename, hash_object=None, chunk_size=1024 * 1024):
    hash_object = hash_object or hashlib.md5()
    if isinstance(hash_object, str):
        hash_object = hashlib.new(hash_object)
    
    with open(filename, "rb") as f:
        while True:
            chunk = f.read(chunk_size)
            if not chunk:
                break
            hash_object.update(chunk)
    return hash_object.hexdigest()

    


import numpy as np
from .boxes_utils import assert_and_normalize_shape


def flip_boxes(boxes, x_center=0, y_center=0, direction='h'):
    """
    Args:
        boxes: (N, 4+K)
        x_center: array-like whose shape is (), (1,), (N,), (1, 1) or (N, 1)
        y_center: array-like whose shape is (), (1,), (N,), (1, 1) or (N, 1)
        direction: str
    """
    assert direction in ['x', 'h', 'horizontal',
                         'y', 'v', 'vertical', 
                         'o', 'b', 'both']
    boxes = np.asarray(boxes, np.float32)
    ret_boxes = boxes.copy()
    
    x_center = np.asarray(x_center, np.float32)
    y_center = np.asarray(y_center, np.float32)
    x_center = assert_and_normalize_shape(x_center, boxes.shape[0])
    y_center = assert_and_normalize_shape(y_center, boxes.shape[0])
    
    if direction in ['o', 'b', 'both', 'x', 'h', 'horizontal']:
        ret_boxes[:, 0] = 2 * x_center - boxes[:, 2] 
        ret_boxes[:, 2] = 2 * x_center - boxes[:, 0]
    if direction in ['o', 'b', 'both', 'y', 'v', 'vertical']:
        ret_boxes[:, 1] = 2 * y_center - boxes[:, 3]
        ret_boxes[:, 3] = 2 * y_center - boxes[:, 1]
    return ret_boxes
    
    
def fliplr_boxes(boxes, x_center=0, y_center=0):
    """
    Args:
        boxes: (N, 4+K)
        x_center: array-like whose shape is (), (1,), (N,), (1, 1) or (N, 1)
        y_center: array-like whose shape is (), (1,), (N,), (1, 1) or (N, 1)
    """
    boxes = np.asarray(boxes, np.float32)
    ret_boxes = boxes.copy()
    
    x_center = np.asarray(x_center, np.float32)
    y_center = np.asarray(y_center, np.float32)
    x_center = assert_and_normalize_shape(x_center, boxes.shape[0])
    y_center = assert_and_normalize_shape(y_center, boxes.shape[0])
     
    ret_boxes[:, 0] = 2 * x_center - boxes[:, 2] 
    ret_boxes[:, 2] = 2 * x_center - boxes[:, 0]
    return ret_boxes
    
    
def flipud_boxes(boxes, x_center=0, y_center=0):
    """
    Args:
        boxes: (N, 4+K)
        x_center: array-like whose shape is (), (1,), (N,), (1, 1) or (N, 1)
        y_center: array-like whose shape is (), (1,), (N,), (1, 1) or (N, 1)
    """
    boxes = np.asarray(boxes, np.float32)
    ret_boxes = boxes.copy()
    
    x_center = np.asarray(x_center, np.float32)
    y_center = np.asarray(y_center, np.float32)
    x_center = assert_and_normalize_shape(x_center, boxes.shape[0])
    y_center = assert_and_normalize_shape(y_center, boxes.shape[0])
    
    ret_boxes[:, 1] = 2 * y_center - boxes[:, 3]
    ret_boxes[:, 3] = 2 * y_center - boxes[:, 1]
    return ret_boxes
    
    
def transpose_boxes(boxes, x_center=0, y_center=0):
    """
    Args:
        boxes: (N, 4+K)
        x_center: array-like whose shape is (), (1,), (N,), (1, 1) or (N, 1)
        y_center: array-like whose shape is (), (1,), (N,), (1, 1) or (N, 1)
    """
    boxes = np.asarray(boxes, np.float32)
    ret_boxes = boxes.copy()
    
    x_center = np.asarray(x_center, np.float32)
    y_center = np.asarray(y_center, np.float32)
    x_center = assert_and_normalize_shape(x_center, boxes.shape[0])
    y_center = assert_and_normalize_shape(y_center, boxes.shape[0])
    
    shift = x_center - y_center
    ret_boxes[:, 0] = boxes[:, 1] + shift
    ret_boxes[:, 1] = boxes[:, 0] - shift
    ret_boxes[:, 2] = boxes[:, 3] + shift
    ret_boxes[:, 3] = boxes[:, 2] - shift
    return ret_boxes


def flip_boxes_in_image(boxes, image_width, image_height, direction='h'):
    """
    Args:
        boxes: (N, 4+K)
        image_width: int
        image_width: int
        direction: str
        
    References:
        `core.bbox.bbox_flip` in mmdetection
        `datasets.pipelines.RandomFlip.bbox_flip` in mmdetection
    """
    x_center = (image_width - 1) * 0.5
    y_center = (image_height - 1) * 0.5
    ret_boxes = flip_boxes(boxes, x_center, y_center, direction=direction)
    return ret_boxes
    
    
def rot90_boxes_in_image(boxes, image_width, image_height, n=1):
    """Rotate boxes counter-clockwise by 90 degrees.
    
    References:
        np.rot90
        cv2.rotate
        tf.image.rot90
    """
    n = n % 4
    if n == 0:
        ret_boxes = boxes.copy()
    elif n == 1:
        ret_boxes = transpose_boxes(boxes)
        ret_boxes = flip_boxes_in_image(ret_boxes, image_width, image_height, 'v')
    elif n == 2:
        ret_boxes = flip_boxes_in_image(boxes, image_width, image_height, 'o')
    else:
        ret_boxes = transpose_boxes(boxes)
        ret_boxes = flip_boxes_in_image(ret_boxes, image_width, image_height, 'h');
    return ret_boxes
    
    


from typing import Optional

import numpy as np

from .classifier import *
from .detector import *
from .extractor import *

import khandy

__all__ = ['extract_feature']


def extract_feature(image: np.ndarray, extractor: Extractor, 
                    det_objects: Optional[DetObjects] = None,
                    detector: Optional[BaseDetector] = None,
                    **detector_kwargs) -> DetObjects:
    if det_objects is not None:
        pass
    elif detector is not None:
        det_objects = detector(image, **detector_kwargs)
    else:
        det_objects.feats = np.expand_dims(extractor(image), axis=0)
        return det_objects
    
    feats = np.empty((len(det_objects), extractor.feature_dim), dtype=np.float32)
    for det_obj in det_objects:
        cropped = khandy.crop_image(
            image, round(det_obj.x_min), round(det_obj.y_min),
            round(det_obj.x_max), round(det_obj.y_max))
        feats.append(extractor(cropped))
    det_objects.feats = feats
    return det_objects




__all__ = ['letterbox_2d_points', 'unletterbox_2d_points']


def letterbox_2d_points(points, lb_detail, copy=True):
    if copy:
        points = points.copy()
    points[..., 0::2] = points[..., 0::2] * lb_detail.x_scale + lb_detail.pad_left
    points[..., 1::2] = points[..., 1::2] * lb_detail.y_scale + lb_detail.pad_top
    return points


def unletterbox_2d_points(points, lb_detail, copy=True):
    if copy:
        points = points.copy()
        
    points[..., 0::2] = (points[..., 0::2] - lb_detail.pad_left) / lb_detail.x_scale
    points[..., 1::2] = (points[..., 1::2] - lb_detail.pad_top) / lb_detail.y_scale
    return points
    



import numpy as np
from .boxes_utils import assert_and_normalize_shape


def rotate_boxes(boxes, angle, x_center=0, y_center=0, scale=1, 
                 degrees=True, return_rotated_boxes=False):
    """
    Args:
        boxes: (N, 4+K)
        angle: array-like whose shape is (), (1,), (N,), (1, 1) or (N, 1)
        x_center: array-like whose shape is (), (1,), (N,), (1, 1) or (N, 1)
        y_center: array-like whose shape is (), (1,), (N,), (1, 1) or (N, 1)
        scale: array-like whose shape is (), (1,), (N,), (1, 1) or (N, 1)
            scale factor in x and y dimension
        degrees: bool
        return_rotated_boxes: bool
    """
    boxes = np.asarray(boxes, np.float32)
    
    angle = np.asarray(angle, np.float32)
    x_center = np.asarray(x_center, np.float32)
    y_center = np.asarray(y_center, np.float32)
    scale = np.asarray(scale, np.float32)
    
    angle = assert_and_normalize_shape(angle, boxes.shape[0])
    x_center = assert_and_normalize_shape(x_center, boxes.shape[0])
    y_center = assert_and_normalize_shape(y_center, boxes.shape[0])
    scale = assert_and_normalize_shape(scale, boxes.shape[0])
    
    if degrees:
        angle = np.deg2rad(angle)
    cos_val = scale * np.cos(angle)
    sin_val = scale * np.sin(angle)
    x_shift = x_center - x_center * cos_val + y_center * sin_val
    y_shift = y_center - x_center * sin_val - y_center * cos_val
    
    x_mins, y_mins = boxes[:,0], boxes[:,1]
    x_maxs, y_maxs = boxes[:,2], boxes[:,3]
    x00 = x_mins * cos_val - y_mins * sin_val + x_shift
    x10 = x_maxs * cos_val - y_mins * sin_val + x_shift
    x11 = x_maxs * cos_val - y_maxs * sin_val + x_shift
    x01 = x_mins * cos_val - y_maxs * sin_val + x_shift
    
    y00 = x_mins * sin_val + y_mins * cos_val + y_shift
    y10 = x_maxs * sin_val + y_mins * cos_val + y_shift
    y11 = x_maxs * sin_val + y_maxs * cos_val + y_shift
    y01 = x_mins * sin_val + y_maxs * cos_val + y_shift
    
    rotated_boxes = np.stack([x00, y00, x10, y10, x11, y11, x01, y01], axis=-1)
    ret_x_mins = np.min(rotated_boxes[:,0::2], axis=1)
    ret_y_mins = np.min(rotated_boxes[:,1::2], axis=1)
    ret_x_maxs = np.max(rotated_boxes[:,0::2], axis=1)
    ret_y_maxs = np.max(rotated_boxes[:,1::2], axis=1)
    
    if boxes.ndim == 4:
        ret_boxes = np.stack([ret_x_mins, ret_y_mins, ret_x_maxs, ret_y_maxs], axis=-1)
    else:
        ret_boxes = boxes.copy()
        ret_boxes[:, :4] = np.stack([ret_x_mins, ret_y_mins, ret_x_maxs, ret_y_maxs], axis=-1)
        
    if not return_rotated_boxes:
        return ret_boxes
    else:
        return ret_boxes, rotated_boxes
    
    
def rotate_boxes_wrt_centers(boxes, angle, scale=1, degrees=True,  
                             return_rotated_boxes=False):
    """
    Args:
        boxes: (N, 4+K)
        angle: array-like whose shape is (), (1,), (N,), (1, 1) or (N, 1)
        scale: array-like whose shape is (), (1,), (N,), (1, 1) or (N, 1)
            scale factor in x and y dimension
        degrees: bool
        return_rotated_boxes: bool
    """
    boxes = np.asarray(boxes, np.float32)
    
    angle = np.asarray(angle, np.float32)
    scale = np.asarray(scale, np.float32)
    angle = assert_and_normalize_shape(angle, boxes.shape[0])
    scale = assert_and_normalize_shape(scale, boxes.shape[0])
    
    if degrees:
        angle = np.deg2rad(angle)
    cos_val = scale * np.cos(angle)
    sin_val = scale * np.sin(angle)
    
    x_centers = boxes[:, 2] + boxes[:, 0]
    y_centers = boxes[:, 3] + boxes[:, 1]
    x_centers *= 0.5
    y_centers *= 0.5
    
    half_widths = boxes[:, 2] - boxes[:, 0]
    half_heights = boxes[:, 3] - boxes[:, 1]
    half_widths *= 0.5
    half_heights *= 0.5
    
    half_widths_cos = half_widths * cos_val
    half_widths_sin = half_widths * sin_val
    half_heights_cos = half_heights * cos_val
    half_heights_sin = half_heights * sin_val
    
    x00 = -half_widths_cos + half_heights_sin
    x10 = half_widths_cos + half_heights_sin
    x11 = half_widths_cos - half_heights_sin
    x01 = -half_widths_cos - half_heights_sin
    x00 += x_centers
    x10 += x_centers
    x11 += x_centers
    x01 += x_centers
    
    y00 = -half_widths_sin - half_heights_cos
    y10 = half_widths_sin - half_heights_cos
    y11 = half_widths_sin + half_heights_cos
    y01 = -half_widths_sin + half_heights_cos
    y00 += y_centers
    y10 += y_centers
    y11 += y_centers
    y01 += y_centers
    
    rotated_boxes = np.stack([x00, y00, x10, y10, x11, y11, x01, y01], axis=-1)
    ret_x_mins = np.min(rotated_boxes[:,0::2], axis=1)
    ret_y_mins = np.min(rotated_boxes[:,1::2], axis=1)
    ret_x_maxs = np.max(rotated_boxes[:,0::2], axis=1)
    ret_y_maxs = np.max(rotated_boxes[:,1::2], axis=1)
    
    if boxes.ndim == 4:
        ret_boxes = np.stack([ret_x_mins, ret_y_mins, ret_x_maxs, ret_y_maxs], axis=-1)
    else:
        ret_boxes = boxes.copy()
        ret_boxes[:, :4] = np.stack([ret_x_mins, ret_y_mins, ret_x_maxs, ret_y_maxs], axis=-1)
        
    if not return_rotated_boxes:
        return ret_boxes
    else:
        return ret_boxes, rotated_boxes
    
    


import time

import khandy
import unittest


class TestBenchmark(unittest.TestCase):
    def test_benchmark_function(self):
        def sample_function(x):
            time.sleep(0.1) 
            return x * 2

        output, stats = khandy.benchmark(sample_function, args=(5,), num_repeats=5, num_repeats_burn_in=1)
        self.assertEqual(output, 10)  # 5 * 2
        self.assertIn('mean', stats)
        self.assertIn('stddev', stats)
        self.assertIn('max', stats)
        self.assertIn('min', stats)

    def test_benchmark_with_no_repeats(self):
        with self.assertRaises(AssertionError):
            khandy.benchmark(lambda x: x, num_repeats=-1)
        with self.assertRaises(AssertionError):
            khandy.benchmark(lambda x: x, num_repeats=3.5)
            

if __name__ == '__main__':
    unittest.main()
    
    


import logging
import os
import re
import shutil
import warnings
from datetime import datetime
from typing import Callable, List, Literal, Optional, Union


def get_path_stem(path: str) -> str:
    """
    References:
        `std::filesystem::path::stem` since C++17
    """
    return os.path.splitext(os.path.basename(path))[0]


def replace_path_stem(path: str, new_stem: Union[str, Callable[[str], str]]) -> str:
    dirname, basename = os.path.split(path)
    stem, extension = os.path.splitext(basename)
    if isinstance(new_stem, str):
        return os.path.join(dirname, new_stem + extension)
    elif callable(new_stem):
        return os.path.join(dirname, new_stem(stem) + extension)
    else:
        raise TypeError('Unsupported Type!')
        

def get_path_extension(path: str) -> str:
    """
    References:
        `std::filesystem::path::extension` since C++17
        
    Notes:
        Not fully consistent with `std::filesystem::path::extension`
    """
    return os.path.splitext(os.path.basename(path))[1]
    

def replace_path_extension(path: str, new_extension: Optional[str] = None) -> str:
    """Replaces the extension with new_extension or removes it when the default value is used.
    Firstly, if this path has an extension, it is removed. Then, a dot character is appended 
    to the pathname, if new_extension is not empty or does not begin with a dot character.

    References:
        `std::filesystem::path::replace_extension` since C++17
    """
    path_wo_ext = os.path.splitext(path)[0]
    if new_extension == '' or new_extension is None:
        return path_wo_ext
    elif new_extension.startswith('.'):
        return f'{path_wo_ext}{new_extension}'
    else:
        return f'{path_wo_ext}.{new_extension}'
    
    
def normalize_extension(extension: str) -> str:
    if extension.startswith('.'):
        new_extension = extension.lower()
    else:
        new_extension =  '.' + extension.lower()
    return new_extension


def is_path_in_extensions(path: str, extensions: Union[str, List[str]]) -> bool:
    if isinstance(extensions, str):
        extensions = [extensions]
    extensions = [normalize_extension(item) for item in extensions]
    extension = get_path_extension(path)
    return extension.lower() in extensions


def get_path_parts(path: str, sep: Optional[Literal['/', '\\']] = None) -> List[str]:
    """Splits a file path into its constituent parts.  
  
    Args:
        path (str): The file path to be split.  
        sep (Optional[Literal['/', '\\']], optional): The separator to use for splitting the path.  
            Must be None, '/', or '\\'. Defaults to the system-specific path separator (os.path.sep).  
  
    Returns:
        List[str]: A list of strings representing the individual parts of the path.  
  
    Raises:
        ValueError: If sep is not None, '/', or '\\'.

    Notes:
        This implementation is different from pathlib.Path.parts. 
    """
    if sep is not None and sep not in ('/', '\\'):
        raise ValueError("sep must be None, '/', or '\\'")
    sep = sep or os.path.sep
    path = os.path.normpath(path)
    parts = path.split(sep)
    return parts


def normalize_path(path: str, norm_case: bool = True) -> str:
    """
    References:
        https://en.cppreference.com/w/cpp/filesystem/canonical
    """
    # On Unix and Windows, return the argument with an initial 
    # component of ~ or ~user replaced by that user's home directory.
    path = os.path.expanduser(path)
    # Return a normalized absolutized version of the pathname path. 
    # On most platforms, this is equivalent to calling the function 
    # normpath() as follows: normpath(join(os.getcwd(), path)).
    path = os.path.abspath(path)
    if norm_case:
        # Normalize the case of a pathname. On Windows, 
        # convert all characters in the pathname to lowercase, 
        # and also convert forward slashes to backward slashes. 
        # On other operating systems, return the path unchanged.
        path = os.path.normcase(path)
    return path
    

def upsert_prefix_into_path_stem(filename: str, prefix: str, validator: Optional[Callable[[str], bool]] = None, sep: str = '_') -> str:
    """Upsert a prefix into the first part of the file stem in a given file path.

    Args:
        filename (str): The full path of the file.
        prefix (str): The prefix to be inserted or updated.
        validator (Optional[Callable[[str], bool]]): An optional callable that takes a string and returns a boolean.
            If provided, it will be used to determine if the prefix should be inserted or updated.
            If the validator returns False for the first part of the stem, the prefix will be inserted.
            If the validator returns True, the first part of the stem will be replaced with the prefix.
        sep (str, optional): The separator used to split the stem. Defaults to '_'.

    Returns:
        str: The new file path with the updated stem.

    Raises:
        AssertionError: If the validator is not None and is not callable.
    """
    assert validator is None or callable(validator)
    dirname, basename = os.path.split(filename)
    stem, extension = os.path.splitext(basename)
    if stem.startswith('.') and extension == '':
        stem, extension = extension, stem

    stem_parts = stem.split(sep, maxsplit=1)
    first_part = stem_parts[0]
    if first_part == '':
        stem_parts = stem_parts[1:]
        
    if validator is None:
        stem_parts.insert(0, prefix)
    elif not validator(first_part):
        stem_parts.insert(0, prefix)
    else:
        if len(stem_parts) != 0:
            stem_parts[0] = prefix
    new_basename = sep.join(stem_parts) + extension
    return os.path.join(dirname, new_basename)


def upsert_suffix_into_path_stem(filename: str, suffix: str, validator: Optional[Callable[[str], bool]] = None, sep: str = '_') -> str:
    """Upsert a suffix into the last part of the file stem in a given file path.

    Args:
        filename (str): The full path of the file.
        suffix (str): The suffix to be inserted or updated.
        validator (Optional[Callable[[str], bool]]): An optional callable function that takes a string as input
            and returns a boolean. If provided, it will be used to determine if the suffix should be inserted or updated.
            If the validator returns False for the last part of the stem, the suffix will be appended.
            If the validator returns True, the last part of the stem will be replaced with the suffix.
        sep (str, optional): The separator used to split the stem. Defaults to '_'.

    Returns:
        str: The new file path with the updated stem.

    Raises:
        AssertionError: If the validator is not None and is not callable.
    """
    assert validator is None or callable(validator)
    dirname, basename = os.path.split(filename)
    stem, extension = os.path.splitext(basename)
    if stem.startswith('.') and extension == '':
        stem, extension = extension, stem
        
    stem_parts = stem.rsplit(sep, maxsplit=1)
    last_part = stem_parts[-1]
    if last_part == '':
        stem_parts = stem_parts[:-1]
    
    if validator is None:
        stem_parts.append(suffix)
    elif not validator(last_part):
        stem_parts.append(suffix)
    else:
        if len(stem_parts) != 0:
            stem_parts[-1] = suffix
    new_basename = sep.join(stem_parts) + extension
    return os.path.join(dirname, new_basename)
    

def makedirs(name, mode=0o755):
    """
    References:
        mmcv.mkdir_or_exist
    """
    warnings.warn('`makedirs` will be deprecated!')
    if name == '':
        return
    name = os.path.expanduser(name)
    os.makedirs(name, mode=mode, exist_ok=True)


def listdirs(paths, path_sep=None, full_path=True):
    """Enhancement on `os.listdir`
    """
    warnings.warn('`listdirs` will be deprecated!')
    assert isinstance(paths, (str, tuple, list))
    if isinstance(paths, str):
        path_sep = path_sep or os.path.pathsep
        paths = paths.split(path_sep)
        
    all_filenames = []
    for path in paths:
        path_ex = os.path.expanduser(path)
        filenames = os.listdir(path_ex)
        if full_path:
            filenames = [os.path.join(path_ex, filename) for filename in filenames]
        all_filenames.extend(filenames)
    return all_filenames


def get_all_filenames(path, extensions=None, is_valid_file=None):
    warnings.warn('`get_all_filenames` will be deprecated, use `list_files_in_dir` with `recursive=True` instead!')
    if (extensions is not None) and (is_valid_file is not None):
        raise ValueError("Both extensions and is_valid_file cannot "
                         "be not None at the same time")
    if is_valid_file is None:
        if extensions is not None:
            def is_valid_file(filename):
                return is_path_in_extensions(filename, extensions)
        else:
            def is_valid_file(filename):
                return True

    all_filenames = []
    path_ex = os.path.expanduser(path)
    for root, _, filenames in sorted(os.walk(path_ex, followlinks=True)):
        for filename in sorted(filenames):
            fullname = os.path.join(root, filename)
            if is_valid_file(fullname):
                all_filenames.append(fullname)
    return all_filenames


def get_top_level_dirs(path, full_path=True):
    warnings.warn('`get_top_level_dirs` will be deprecated, use `list_dirs_in_dir` instead!')
    if path is None:
        path = os.getcwd()
    path_ex = os.path.expanduser(path)
    filenames = os.listdir(path_ex)
    if full_path:
        return [os.path.join(path_ex, item) for item in filenames
                if os.path.isdir(os.path.join(path_ex, item))]
    else:
        return [item for item in filenames
                if os.path.isdir(os.path.join(path_ex, item))]


def get_top_level_files(path, full_path=True):
    warnings.warn('`get_top_level_files` will be deprecated, use `list_files_in_dir` instead!')
    if path is None:
        path = os.getcwd()
    path_ex = os.path.expanduser(path)
    filenames = os.listdir(path_ex)
    if full_path:
        return [os.path.join(path_ex, item) for item in filenames
                if os.path.isfile(os.path.join(path_ex, item))]
    else:
        return [item for item in filenames
                if os.path.isfile(os.path.join(path_ex, item))]
                

def list_items_in_dir(path: Optional[str] = None, recursive: bool = False, full_path: bool = True) -> List[str]:
    """List all entries in directory
    """
    if path is None:
        path = os.getcwd()
    path_ex = os.path.expanduser(path)
    if not os.path.exists(path_ex):
        raise FileNotFoundError(f'{path_ex} is not found!')
    elif not os.path.isdir(path_ex):
        raise NotADirectoryError(f'{path_ex} is not a directory!')
    
    if not recursive:
        names = os.listdir(path_ex)
        if full_path:
            return [os.path.join(path_ex, name) for name in sorted(names)]
        else:
            return sorted(names)
    else:
        all_names = []
        for root, dirnames, filenames in sorted(os.walk(path_ex, followlinks=True)):
            all_names += [os.path.join(root, name) for name in sorted(dirnames)]
            all_names += [os.path.join(root, name) for name in sorted(filenames)]
        return all_names


def list_dirs_in_dir(path: Optional[str] = None, recursive: bool = False, full_path: bool = True) -> List[str]:
    """List all dirs in directory
    """
    if path is None:
        path = os.getcwd()
    path_ex = os.path.expanduser(path)
    if not os.path.exists(path_ex):
        raise FileNotFoundError(f'{path_ex} is not found!')
    elif not os.path.isdir(path_ex):
        raise NotADirectoryError(f'{path_ex} is not a directory!')
    
    if not recursive:
        with os.scandir(path_ex) as it:
            if full_path:
                return sorted([entry.path for entry in it if entry.is_dir()])
            else:
                return sorted([entry.name for entry in it if entry.is_dir()])
    else:
        all_names = []
        for root, dirnames, _ in sorted(os.walk(path_ex, followlinks=True)):
            all_names += [os.path.join(root, name) for name in sorted(dirnames)]
        return all_names


def list_files_in_dir(path: Optional[str] = None, recursive: bool = False, full_path: bool = True) -> List[str]:
    """List all files in directory
    """
    if path is None:
        path = os.getcwd()
    path_ex = os.path.expanduser(path)
    if not os.path.exists(path_ex):
        raise FileNotFoundError(f'{path_ex} is not found!')
    elif not os.path.isdir(path_ex):
        raise NotADirectoryError(f'{path_ex} is not a directory!')
    
    if not recursive:
        with os.scandir(path_ex) as it:
            if full_path:
                return sorted([entry.path for entry in it if entry.is_file()])
            else:
                return sorted([entry.name for entry in it if entry.is_file()])
    else:
        all_names = []
        for root, _, filenames in sorted(os.walk(path_ex, followlinks=True)):
            all_names += [os.path.join(root, name) for name in sorted(filenames)]
        return all_names
        

def get_folder_size(path: str) -> int:
    if path is None:
        path = os.getcwd()
    path_ex = os.path.expanduser(path)
    if not os.path.exists(path_ex):
        raise FileNotFoundError(f'{path_ex} is not found!')
    elif not os.path.isdir(path_ex):
        raise NotADirectoryError(f'{path_ex} is not a directory!')
    
    total_size = 0
    for root, _, filenames in os.walk(path_ex):
        for name in filenames:
            total_size += os.path.getsize(os.path.join(root, name))
    return total_size

    
def escape_filename(filename: str, new_char: str = '_') -> str:
    assert isinstance(new_char, str)
    control_chars = ''.join((map(chr, range(0x00, 0x20))))
    pattern = r'[\\/*?:"<>|{}]'.format(control_chars)
    return re.sub(pattern, new_char, filename)


def replace_invalid_filename_char(filename, new_char='_'):
    warnings.warn('`replace_invalid_filename_char` will be deprecated, use `escape_filename` instead!')
    return escape_filename(filename, new_char)


def copy_file(src: str, dst_dir: str, action_if_exist: Optional[Literal['ignore', 'rename']] = 'rename') -> str:
    """
    Args:
        src: source file path
        dst_dir: dest dir
        action_if_exist: 
            None: same as shutil.copy
            ignore: when dest file exists, don't copy and return None
            rename: when dest file exists, copy after rename
            
    Returns:
        dest filename
    """
    dst = os.path.join(dst_dir, os.path.basename(src))

    if action_if_exist is None:
        os.makedirs(dst_dir, exist_ok=True)
        shutil.copy(src, dst)
    elif action_if_exist.lower() == 'ignore':
        if os.path.exists(dst):
            warnings.warn(f'{dst} already exists, do not copy!')
            return dst
        os.makedirs(dst_dir, exist_ok=True)
        shutil.copy(src, dst)
    elif action_if_exist.lower() == 'rename':
        suffix = 2
        stem, extension = os.path.splitext(os.path.basename(src))
        while os.path.exists(dst):
            dst = os.path.join(dst_dir, f'{stem} ({suffix}){extension}')
            suffix += 1
        os.makedirs(dst_dir, exist_ok=True)
        shutil.copy(src, dst)
    else:
        raise ValueError('Invalid action_if_exist, got {}.'.format(action_if_exist))
        
    return dst
    
    
def move_file(src: str, dst_dir: str, action_if_exist: Optional[Literal['ignore', 'rename']] = 'rename') -> str:
    """
    Args:
        src: source file path
        dst_dir: dest dir
        action_if_exist: 
            None: same as shutil.move
            ignore: when dest file exists, don't move and return None
            rename: when dest file exists, move after rename
            
    Returns:
        dest filename
    """
    dst = os.path.join(dst_dir, os.path.basename(src))

    if action_if_exist is None:
        os.makedirs(dst_dir, exist_ok=True)
        shutil.move(src, dst)
    elif action_if_exist.lower() == 'ignore':
        if os.path.exists(dst):
            warnings.warn(f'{dst} already exists, do not move!')
            return dst
        os.makedirs(dst_dir, exist_ok=True)
        shutil.move(src, dst)
    elif action_if_exist.lower() == 'rename':
        suffix = 2
        stem, extension = os.path.splitext(os.path.basename(src))
        while os.path.exists(dst):
            dst = os.path.join(dst_dir, f'{stem} ({suffix}){extension}')
            suffix += 1
        os.makedirs(dst_dir, exist_ok=True)
        shutil.move(src, dst)
    else:
        raise ValueError('Invalid action_if_exist, got {}.'.format(action_if_exist))
        
    return dst
    
    
def rename_file(src: str, dst: str, action_if_exist: Optional[Literal['ignore', 'rename']] = 'rename') -> str:
    """
    Args:
        src: source file path
        dst: dest file path
        action_if_exist: 
            None: same as os.rename
            ignore: when dest file exists, don't rename and return None
            rename: when dest file exists, rename it
            
    Returns:
        dest filename
    """
    if dst == src:
        return dst
    dst_dir = os.path.dirname(os.path.abspath(dst))
    
    if action_if_exist is None:
        os.makedirs(dst_dir, exist_ok=True)
        os.rename(src, dst)
    elif action_if_exist.lower() == 'ignore':
        if os.path.exists(dst):
            warnings.warn(f'{dst} already exists, do not rename!')
            return dst
        os.makedirs(dst_dir, exist_ok=True)
        os.rename(src, dst)
    elif action_if_exist.lower() == 'rename':
        suffix = 2
        stem, extension = os.path.splitext(os.path.basename(dst))
        while os.path.exists(dst):
            dst = os.path.join(dst_dir, f'{stem} ({suffix}){extension}')
            suffix += 1
        os.makedirs(dst_dir, exist_ok=True)
        os.rename(src, dst)
    else:
        raise ValueError('Invalid action_if_exist, got {}.'.format(action_if_exist))
        
    return dst
    
    
def _get_default_logger(logger_name):
    logger = logging.getLogger(logger_name)
    logger.setLevel(logging.INFO)

    already_have = False
    for handler in logger.handlers[:]:
        if type(handler) == logging.StreamHandler:
            already_have = True
            break
        
    if not already_have:
        console = logging.StreamHandler()
        console.setFormatter(logging.Formatter('%(message)s'))
        logger.addHandler(console)
    return logger


def copy_dir(src_dir: str, dst_dir: str, 
             action_if_file_exist: Optional[Literal['ignore', 'rename']] = 'rename', 
             logger: Optional[logging.Logger] = None):
    """Recursively copies a directory and its contents from the source directory to the destination directory.  
  
    Args:  
        src_dir (str): Path to the source directory to be copied.  
        dst_dir (str): Path to the destination directory where the source directory will be copied.  
        action_if_file_exist (str, optional): Action to be taken if a file already exists in the destination directory.
            Defaults to 'rename'. Other options refers to `khandy.copy_file`.
        logger (logging.Logger, optional): A custom logger to use for logging messages. 
            If not provided, a default logger is used.
  
    Returns:
        None
  
    Raises:  
        KeyboardInterrupt: If the operation is interrupted by the user.  
        Exception: If any other exception occurs during the copying process, the exception will be caught and logged.  
    """
    if logger is None:
        logger = _get_default_logger(__name__)
    try:
        logger.info(f'{datetime.now()} MAKE {src_dir}')
        for root, dirs, files in os.walk(src_dir):
            dst_root = os.path.normpath(os.path.join(dst_dir, os.path.relpath(root, src_dir)))
            for file in files:
                copy_file(os.path.join(root, file), dst_root, action_if_file_exist)
                logger.info(f'{datetime.now()} COPY {os.path.join(root, file)} -> {dst_root}')
            for dir in dirs:
                os.makedirs(os.path.join(dst_root, dir), exist_ok=True)
                logger.info(f'{datetime.now()} MAKE {os.path.join(root, dir)}')
    except KeyboardInterrupt as e:
        raise e
    except Exception as e:
        logger.error(f'{datetime.now()} ERROR: {e}')


def move_dir(src_dir: str, dst_dir: str, 
             action_if_file_exist: Optional[Literal['ignore', 'rename']] = 'rename', 
             rmtree_after_move: bool = False, logger: Optional[logging.Logger] = None):
    """Recursively moves a directory and its contents from the source directory to the destination directory.  
  
    Args:  
        src_dir (str): Path to the source directory to be moved.  
        dst_dir (str): Path to the destination directory where the source directory will be moved.  
        action_if_file_exist (str, optional): Action to be taken if a file already exists in the destination directory.  
            Defaults to 'rename'. Other options refers to `khandy.move_file`.
        rmtree_after_move (bool, optional): Whether remove src dir after move.
        logger (logging.Logger, optional): A custom logger to use for logging messages. 
            If not provided, a default logger is used. 
  
    Returns:  
        None
  
    Raises:  
        KeyboardInterrupt: If the operation is interrupted by the user.  
        Exception: If any other exception occurs during the moving process, the exception will be caught and logged.  
    """
    if logger is None:
        logger = _get_default_logger(__name__)
    try:
        for root, dirs, files in os.walk(src_dir):
            dst_root = os.path.normpath(os.path.join(dst_dir, os.path.relpath(root, src_dir)))
            for file in files:
                move_file(os.path.join(root, file), dst_root, action_if_file_exist)
                logger.info(f'{datetime.now()} MOVE {os.path.join(root, file)} -> {dst_root}')
            for dir in dirs:
                os.makedirs(os.path.join(dst_root, dir), exist_ok=True)
        if rmtree_after_move:
            shutil.rmtree(src_dir)
    except KeyboardInterrupt as e:
        raise e
    except Exception as e:
        logger.error(f'{datetime.now()} ERROR: {e}')
        

def rename_files(old_names: List[str], new_names: List[str], 
                 action_if_file_exist: Optional[Literal['ignore', 'rename']] = 'rename', 
                 logger: Optional[logging.Logger] = None):
    """Renames a list of files based on provided old and new file name pairs.  
    
    Args:
        old_names (List[str]): A list of old file names to be renamed.  
        new_names (List[str]): A list of new file names corresponding to the old file names.  
        action_if_file_exist (str, optional): Specifies the action to be taken if a new file already exists.  
            Defaults to 'rename'. Other options refers to `khandy.rename_file`.
        logger (logging.Logger, optional): A custom logger for recording log messages. 
            If not provided, a default logger is used. 
  
    Returns:
        None
  
    Raises:
        KeyboardInterrupt: If the operation is interrupted by the user.  
        Exception: If any other exception occurs during the renaming process, the exception will be caught and logged.  
    """
    if logger is None:
        logger = _get_default_logger(__name__)
    try:
        for k, (old_name, new_name) in enumerate(zip(old_names, new_names)):
            if new_name == old_name:
                logger.info(f'{datetime.now()} [{k+1}/{len(old_names)}] KEEP SAME: {old_name}')
                continue
            new_name = rename_file(old_name, new_name, action_if_file_exist)
            logger.info(f'{datetime.now()} [{k+1}/{len(old_names)}] RENAME {old_name} -> {new_name}')
    except KeyboardInterrupt as e:
        raise e
    except Exception as e:
        logger.error(f'{datetime.now()} ERROR: {e}')



import base64
import json
import numbers
import os
import pickle
import warnings
from collections import OrderedDict
from typing import List


def load_list(filename, encoding='utf-8', start=0, stop=None) -> List[str]:
    assert isinstance(start, numbers.Integral) and start >= 0
    assert (stop is None) or (isinstance(stop, numbers.Integral) and stop > start)
    
    lines = []
    with open(filename, 'r', encoding=encoding) as f:
        for _ in range(start):
            f.readline()
        for k, line in enumerate(f):
            if (stop is not None) and (k + start > stop):
                break
            lines.append(line.rstrip('\n'))
    return lines


def save_list(filename, list_obj, encoding='utf-8', append_break=True):
    with open(filename, 'w', encoding=encoding) as f:
        if append_break:
            for item in list_obj:
                f.write(str(item) + '\n')
        else:
            for item in list_obj:
                f.write(str(item))


def load_json(filename, encoding='utf-8'):
    with open(filename, 'r', encoding=encoding) as f:
        data = json.load(f, object_pairs_hook=OrderedDict)
    return data


def save_json(filename, data, encoding='utf-8', indent=4, cls=None, sort_keys=False):
    if not filename.endswith('.json'):
        filename = filename + '.json'
    with open(filename, 'w', encoding=encoding) as f:
        json.dump(data, f, indent=indent, separators=(',',': '),
                  ensure_ascii=False, cls=cls, sort_keys=sort_keys)


def load_bytes(filename, use_base64: bool = False) -> bytes:
    """Open the file in bytes mode, read it, and close the file.
    
    References:
        pathlib.Path.read_bytes
    """
    with open(filename, 'rb') as f:
        data = f.read()
    if use_base64:
        data = base64.b64encode(data)
    return data


def save_bytes(filename, data: bytes, use_base64: bool = False) -> int:
    """Open the file in bytes mode, write to it, and close the file.
    
    References:
        pathlib.Path.write_bytes
    """
    if use_base64:
        data = base64.b64decode(data)
    with open(filename, 'wb') as f:
        ret = f.write(data)
    return ret


def load_as_base64(filename) -> bytes:
    warnings.warn('khandy.load_as_base64 will be deprecated, use khandy.load_bytes instead!')
    return load_bytes(filename, True)


def load_object(filename):
    with open(filename, 'rb') as f:
        return pickle.load(f)
        
        
def save_object(filename, obj):
    with open(filename, 'wb') as f:
        pickle.dump(obj, f)
        

def get_file_header(file_or_buffer, num_bytes=32):
    """Retrieve the file header of a given file or buffer.

    Args:
        file_or_buffer: A file path or buffer object.
        num_bytes: The number of bytes to read from the beginning of the 
            file or buffer. Defaults to 32.

    Returns:
        A bytes or bytearray object representing the retrieved file header.

    Raises:
        TypeError: If the input type is not supported as a file input.
    """
    if isinstance(file_or_buffer, (str, os.PathLike)):
        with open(file_or_buffer, 'rb') as f:
            file_header = f.read(num_bytes)
    elif (hasattr(file_or_buffer, 'read') and 
          hasattr(file_or_buffer, 'tell') and 
          hasattr(file_or_buffer, 'seek')):
        location = file_or_buffer.tell()
        file_header = file_or_buffer.read(num_bytes)
        file_or_buffer.seek(location)
    elif isinstance(file_or_buffer, memoryview):
        return bytearray(file_or_buffer[:num_bytes].tolist())
    elif isinstance(file_or_buffer, (bytes, bytearray)):
        file_header = file_or_buffer[:num_bytes]
    else:
        raise TypeError('Unsupported type as file input: %s' % type(file_or_buffer))
    return file_header




import numpy as np

__all__ = ['scale_2d_points']


def scale_2d_points(points, x_scale=1, y_scale=1, x_center=0, y_center=0, copy=True):
    """Scale 2d points.
    
    Args:
        points: (..., 2N)
        x_scale: scale factor in x dimension
        y_scale: scale factor in y dimension
        x_center: scale center in x dimension
        y_center: scale center in y dimension
    """
    points = np.array(points, dtype=np.float32, copy=copy)
    x_scale = np.asarray(x_scale, np.float32)
    y_scale = np.asarray(y_scale, np.float32)
    x_center = np.asarray(x_center, np.float32)
    y_center = np.asarray(y_center, np.float32)
    
    x_shift = 1 - x_scale
    y_shift = 1 - y_scale
    x_shift *= x_center
    y_shift *= y_center
    
    points[..., 0::2] *= x_scale
    points[..., 1::2] *= y_scale
    points[..., 0::2] += x_shift
    points[..., 1::2] += y_shift
    return points
    
    



import re
from dataclasses import dataclass
from enum import Enum
from typing import List, Optional, Tuple

CONTENT_WITH_HW_PAREN_PATTERN = r'(?:(?P<out_paren>[^(]+))?'
CONTENT_WITH_HW_PAREN_PATTERN += r'(?:[(](?P<in_paren>[^)]*)[)])?'
CONTENT_WITH_HW_PAREN_PATTERN_OBJ = re.compile(CONTENT_WITH_HW_PAREN_PATTERN)

CONTENT_WITH_FW_PAREN_PATTERN = r'(?:(?P<out_paren>[^（]+))?'
CONTENT_WITH_FW_PAREN_PATTERN += r'(?:（(?P<in_paren>[^）]*)）)?'
CONTENT_WITH_FW_PAREN_PATTERN_OBJ = re.compile(CONTENT_WITH_FW_PAREN_PATTERN)

CONTENT_IN_PAREN_PATTERN = r"\([^)]*\)|（[^）]*）"
CONTENT_IN_PAREN_PATTERN_OBJ = re.compile(CONTENT_IN_PAREN_PATTERN)


def has_nested_or_unmatched_paren(string: str, paren_type: str = 'hw') -> bool:
    """Check if a string contains nested or unmatched parentheses.
  
    Args:
        string (str): The input string to be checked.
        paren_type (str): The type of parentheses to be used. Options are 'fw' (full-width) and 'hw' (half-width). Defaults to 'hw'.
  
    Returns:
        bool: True if the string contains nested or unmatched parentheses, False otherwise.
  
    Raises:
        ValueError: If paren_type is not one of 'fw' or 'hw'.
    """  
    if paren_type == 'fw': # full-width
        left_char, right_char = '（）'
    elif paren_type == 'hw': # half-width
        left_char, right_char = '()'
    else:
        raise ValueError(f'paren_type only support fw and hw, got {paren_type}')
    
    stack = []
    for char in string:
        if char == left_char:
            # If stack is not empty, we have a nested parenthesis
            if len(stack) != 0:  
                return True
            stack.append(char)
        elif char == right_char:
            # If stack is empty, we have an unmatched parenthesis
            if len(stack) == 0:  
                return True
            stack.pop()
    # If stack is not empty, we have unmatched opening parentheses
    return len(stack) != 0


def split_content_with_paren(string: str, paren_type: str = 'hw') -> Tuple[Optional[str], Optional[str]]:
    """Split a string into two parts based on the presence of parentheses of a specified type.
  
    Args:
        string (str): The input string to be split.
        paren_type (str, optional): The type of parentheses to split on. Defaults to 'hw' (half-width parentheses).
            Accepted values are 'hw' for half-width parentheses and 'fw' for full-width parentheses.
  
    Returns:
        Tuple[Optional[str], Optional[str]]: A tuple containing two optional string elements.
            The first element is the part outside the parentheses (or None if not found).
            The second element is the part inside the parentheses (or None if not found).
  
    Raises:  
        AssertionError: If the `paren_type` is not one of 'hw' or 'fw'.
    """ 
    assert paren_type in ('hw', 'fw'), f"Paren type must be either 'hw' or 'fw', got {paren_type}."
    pattern_obj = {
        'hw': CONTENT_WITH_HW_PAREN_PATTERN_OBJ,
        'fw': CONTENT_WITH_FW_PAREN_PATTERN_OBJ
    }[paren_type]

    if has_nested_or_unmatched_paren(string, 'hw'):
        return None, None
    matched = pattern_obj.fullmatch(string)
    if matched is None:
        return None, None
    outside, inside = matched.groups()
    return outside, inside
    

def strip_content_in_paren(string: str) -> str:
    """
    Notes:
        strip_content_in_paren cannot process nested paren correctly
    """
    return re.sub(CONTENT_IN_PAREN_PATTERN_OBJ, "", string)


def is_chinese_char(uchar: str) -> bool:
    """Whether the input char is a Chinese character.

    Args:
        uchar: input char in unicode

    References:
        `is_chinese_char` in https://github.com/thunlp/OpenNRE/
    """
    codepoint = ord(uchar)
    if ((0x4E00 <= codepoint <= 0x9FFF) or # CJK Unified Ideographs
        (0x3400 <= codepoint <= 0x4DBF) or # CJK Unified Ideographs Extension A
        (0xF900 <= codepoint <= 0xFAFF) or # CJK Compatibility Ideographs
        (0x20000 <= codepoint <= 0x2A6DF) or # CJK Unified Ideographs Extension B
        (0x2A700 <= codepoint <= 0x2B73F) or
        (0x2B740 <= codepoint <= 0x2B81F) or
        (0x2B820 <= codepoint <= 0x2CEAF) or
        (0x2F800 <= codepoint <= 0x2FA1F)): # CJK Compatibility Supplement
        return True
    return False


def strip_blank_lines(lines: List[str]) -> List[str]:
    """Strip leading and trailing blank lines from a list of lines.

    Args:
        lines (list[str]): A list of lines to process.

    Returns:
        list[str]: A new list containing the lines with leading and trailing blank lines removed.
    """
    start_index, end_index = None, None
    for k, line in enumerate(lines):
        if line.strip() != '':
            start_index = k
            break
    if start_index is None:
        return []
    for k, line in enumerate(lines[::-1]):
        if line.strip() != '':
            end_index = len(lines) - k
            break
    return lines[start_index: end_index]


class MarkdownTableAlignType(Enum):
    DEFAULT = 0
    LEFT = 1
    RIGHT = 2
    BOTH = 3
    
    
@dataclass
class MarkdownTable:
    headers: List[str]
    align_types: List[MarkdownTableAlignType]
    rows: List[List[str]]
    
    
def _get_cell_align_type(cell: str) -> MarkdownTableAlignType:
    cell = re.sub(r'\s', '', cell)
    cell = re.sub(r'-{2,}', '-', cell)
    if cell == '-':
        return MarkdownTableAlignType.DEFAULT
    elif cell == ':-':
        return MarkdownTableAlignType.LEFT
    elif cell == '-:':
        return MarkdownTableAlignType.RIGHT
    elif cell == ':-:':
        return MarkdownTableAlignType.BOTH
    else:
        raise Exception('parse align type')
    
    
def _split_table_line(line: str, length: Optional[int] = None) -> List[str]:
    line = line.strip('|')
    cells = line.split('|')
    cells = [cell.strip() for cell in cells]
    if length is not None and length > len(cells):
        cells += ['' for _ in range(length - len(cells))]
    return cells


def parse_markdown_table(lines: List[str]) -> MarkdownTable:
    """Parse a markdown table from a list of lines.
  
    Args:
        lines (List[str]): A list of lines containing the markdown table.
  
    Returns:
        MarkdownTable: A parsed markdown table object.
  
    Raises:
        Exception: If a blank line is encountered.
        Exception: If the number of lines is too short.
        Exception: If the number of cells in the header line does not match the number of cells in the splitter line.
        Exception: If the number of cells in a row does not match the number of cells in the header line.
    """
    lines = strip_blank_lines(lines)
    for line in lines:
        if line.strip() == '':
            raise Exception('blank line')
    if len(lines) < 2:
        raise Exception('#line too short')

    headers = _split_table_line(lines[0])
    splitters = _split_table_line(lines[1])
    if len(splitters) != len(headers):
        raise Exception('#cell unmatched')
    align_types = [_get_cell_align_type(item) for item in splitters]

    rows = []
    for line in lines[2:]:
        parts = _split_table_line(line, len(headers))
        if len(parts) != len(headers):
            raise Exception('#cell unmatched')
        rows.append(parts)
        
    return MarkdownTable(headers, align_types, rows)
    

def dumps_markdown_table(table: MarkdownTable, align_header: bool = False) -> List[str]:
    """Convert a khandy.MarkdownTable object to a list of strings representing a Markdown table.
  
    Args:
        table (khandy.MarkdownTable): The Markdown table object to be converted.
        align_header (bool, optional): Whether to align the table header to the specified lengths. Defaults to False.
  
    Returns:
        List[str]: A list of strings representing the Markdown table.
    """
    min_length = 5 
    lengths = [max(min_length, len(header)) for header in table.headers]
    for row in table.rows:
        lengths = [max(len(cell), lengths[k]) for k, cell in enumerate(row)]

    if align_header:
        header_cells = [f'{header}'.ljust(lengths[k]) for k, header in enumerate(table.headers)]
    else:
        header_cells = table.headers

    splitter_cells = []
    for k, align_type in enumerate(table.align_types):
        if align_type == MarkdownTableAlignType.DEFAULT:
            cell = f'{"-" * lengths[k]}'
        elif align_type == MarkdownTableAlignType.LEFT:
            cell = f':{"-" * (lengths[k] - 2)}-'
        elif align_type == MarkdownTableAlignType.RIGHT:
            cell = f'-{"-" * (lengths[k] - 2)}:'
        elif align_type == MarkdownTableAlignType.BOTH:
            cell = f':{"-" * (lengths[k] - 2)}:'
        splitter_cells.append(cell)

    cells_list = [header_cells, splitter_cells]
    for row in table.rows:
        cells_list.append([f'{cell}'.ljust(lengths[k]) for k, cell in enumerate(row)])
    dst_lines = [' | '.join(cells) for cells in cells_list]
    dst_lines = [line.strip() for line in dst_lines]
    return dst_lines




import numpy as np
import PIL
from PIL import Image
from PIL import ImageDraw
from PIL import ImageFont
from PIL import ImageColor


def _is_legal_color(color):
    if color is None:
        return True
    if isinstance(color, str):
        return True
    return isinstance(color, (tuple, list)) and len(color) == 3
        

def _normalize_color(color, pil_mode, swap_rgb=False):
    if color is None:
        return color
    if isinstance(color, str):
        color = ImageColor.getrgb(color)
    gray = color[0]
    if swap_rgb:
        color = (color[2], color[1], color[0])
    if pil_mode == 'L':
        color = gray
    return color
    
    
def draw_text(image, text, position, color=(255,0,0), font=None, font_size=15):
    """Draws text on given image.
    
    Args:
        image (ndarray).
        text (str): text to be drawn.
        position (Tuple[int, int]): position where to be drawn.
        color (List[Union[str, Tuple[int, int, int]]]): text color.
        font (str):  A filename or file-like object containing a TrueType font. If the file is not found in this 
            filename, the loader may also search in other directories, such as the `fonts/` directory on Windows
            or `/Library/Fonts/`, `/System/Library/Fonts/` and `~/Library/Fonts/` on macOS.
        font_size (int): The requested font size in points.

    References:
        torchvision.utils.draw_bounding_boxes
    """
    if isinstance(image, np.ndarray):
        # PIL.Image.fromarray fails with uint16 arrays
        # https://github.com/python-pillow/Pillow/issues/1514
        if (image.dtype == np.uint16) and (image.ndim != 2):
            image = (image / 256).astype(np.uint8)
        pil_image = Image.fromarray(image)
    elif isinstance(image, PIL.Image.Image):
        pil_image = image
    else:
        raise TypeError('Unsupported image type!')
    assert pil_image.mode in ['L', 'RGB', 'RGBA']
    
    assert _is_legal_color(color)
    color = _normalize_color(color, pil_image.mode, isinstance(image, np.ndarray))
    
    if font is None:
        font_object = ImageFont.load_default()
    else:
        font_object = ImageFont.truetype(font, size=font_size)
    
    draw = ImageDraw.Draw(pil_image)
    draw.text((position[0], position[1]), text, 
              fill=color, font=font_object)

    if isinstance(image, np.ndarray):
        return np.asarray(pil_image)
    return pil_image


def draw_bounding_boxes(image, boxes, labels=None, colors=None,
                        fill=False, width=1, font=None, font_size=15):
    """Draws bounding boxes on given image.

    Args:
        image (ndarray).
        boxes (ndarray): ndarray of size (N, 4) containing bounding boxes in (xmin, ymin, xmax, ymax) format.
        labels (List[str]): List containing the labels of bounding boxes.
        colors (List[Union[str, Tuple[int, int, int]]]): List containing the colors of bounding boxes or labels.
        fill (bool): If `True` fills the bounding box with specified color.
        width (int): Width of bounding box.
        font (str):  A filename or file-like object containing a TrueType font. If the file is not found in this 
            filename, the loader may also search in other directories, such as the `fonts/` directory on Windows
            or `/Library/Fonts/`, `/System/Library/Fonts/` and `~/Library/Fonts/` on macOS.
        font_size (int): The requested font size in points.

    References:
        torchvision.utils.draw_bounding_boxes
    """
    if isinstance(image, np.ndarray):
        # PIL.Image.fromarray fails with uint16 arrays
        # https://github.com/python-pillow/Pillow/issues/1514
        if (image.dtype == np.uint16) and (image.ndim != 2):
            image = (image / 256).astype(np.uint8)
        pil_image = Image.fromarray(image)
    elif isinstance(image, PIL.Image.Image):
        pil_image = image
    else:
        raise TypeError('Unsupported image type!')
    pil_image = pil_image.convert('RGB')
    
    if font is None:
        font_object = ImageFont.load_default()
    else:
        font_object = ImageFont.truetype(font, size=font_size)

    if fill:
        draw = ImageDraw.Draw(pil_image, "RGBA")
    else:
        draw = ImageDraw.Draw(pil_image)

    for i, bbox in enumerate(boxes):
        if colors is None:
            color = None
        else:
            color = colors[i]
            
        assert _is_legal_color(color)
        color = _normalize_color(color, pil_image.mode, isinstance(image, np.ndarray))
        
        if fill:
            if color is None:
                fill_color = (255, 255, 255, 100)
            elif isinstance(color, str):
                # This will automatically raise Error if rgb cannot be parsed.
                fill_color = ImageColor.getrgb(color) + (100,)
            elif isinstance(color, tuple):
                fill_color = color + (100,)
            # the first argument of ImageDraw.rectangle:
            # in old version only supports [(x0, y0), (x1, y1)]
            # in new version supports either [(x0, y0), (x1, y1)] or [x0, y0, x1, y1]
            draw.rectangle([(bbox[0], bbox[1]), (bbox[2], bbox[3])], width=width, outline=color, fill=fill_color)
        else:
            draw.rectangle([(bbox[0], bbox[1]), (bbox[2], bbox[3])], width=width, outline=color)

        if labels is not None:
            margin = width + 1
            draw.text((bbox[0] + margin, bbox[1] + margin), labels[i], fill=color, font=font_object)

    if isinstance(image, np.ndarray):
        return np.asarray(pil_image)
    return pil_image
    
    
